#include "StdAfx.h"
#include "GLSLManager.h"

vector<ShaderObject*>	GLSLManager::ShaderObjectList;
vector<ShaderProgram*>	GLSLManager::ShaderProgramList;
ShaderProgram*			GLSLManager::CurrentProgram;
ShaderProgram*			GLSLManager::NullProgram;
FILE*					GLSLManager::fp;
string					GLSLManager::base_dir;
string					GLSLManager::shader_dir;
string					GLSLManager::logfile_path;
char					GLSLManager::tmp_current_dir[MAX_PATH];
RenderTexture*			GLSLManager::current_rt;

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

QuadVBO::QuadVBO(float width, float height, float center)
{
	GLSLVertex data[4];
	data[0].x = -.5f; data[0].y =  .5f;
	data[1].x = -.5f; data[1].y = -.5f;
	data[2].x =  .5f; data[2].y = -.5f;
	data[3].x =  .5f; data[3].y =  .5f;

	data[0].u = 0.f; data[0].v = 0.f;
	data[1].u = 0.f; data[1].v = 1.f;
	data[2].u = 1.f; data[2].v = 1.f;
	data[3].u = 1.f; data[3].v = 0.f;

	glGenBuffers(2, BufferId);

	glBindBuffer(GL_ARRAY_BUFFER, BufferId[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(GLSLVertex)*4, data, GL_STATIC_DRAW);

	data[0].x = -.5f; data[0].y = -.5f;
	data[1].x = -.5f; data[1].y =  .5f;
	data[2].x =  .5f; data[2].y =  .5f;
	data[3].x =  .5f; data[3].y = -.5f;

	data[0].u = 0.f; data[0].v = 0.f;
	data[1].u = 0.f; data[1].v = 1.f;
	data[2].u = 1.f; data[2].v = 1.f;
	data[3].u = 1.f; data[3].v = 0.f;

	glBindBuffer(GL_ARRAY_BUFFER, BufferId[1]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(GLSLVertex)*4, data, GL_STATIC_DRAW);

	glBindBuffer(GL_ARRAY_BUFFER, 0);
}

QuadVBO::~QuadVBO()
{
	glDeleteBuffers(2, BufferId);
}

void QuadVBO::Draw(bool flip)
{
	if( flip )
		glBindBuffer(GL_ARRAY_BUFFER, BufferId[0]);
	else
		glBindBuffer(GL_ARRAY_BUFFER, BufferId[1]);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	glVertexPointer(2, GL_FLOAT, sizeof(GLSLVertex), BUFFER_OFFSET( offsetof( GLSLVertex, x ) ));
	glTexCoordPointer(2, GL_FLOAT, sizeof(GLSLVertex), BUFFER_OFFSET( offsetof( GLSLVertex, u ) ));

	glDrawArrays(GL_QUADS, 0, 4);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);

	glBindBuffer(GL_ARRAY_BUFFER, 0);
}

///////////////////////////GLSL


void GLSLManager::Initialize(string &base_path) 
{
	base_dir = base_path;
	shader_dir = base_dir + ToString("\\Shaders\\");
	logfile_path = base_dir + ToString("\\Log\\ShaderLog.txt");

	NullProgram = CurrentProgram = new ShaderProgram(NULL,NULL,NULL,0);
	fp = fopen(logfile_path.c_str(), "w+");

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	current_rt = NULL;
}


void GLSLManager::Shutdown(void)
{
	delete NullProgram;

	for(unsigned int i=0; i<ShaderObjectList.size(); i++) 
	{
		delete ShaderObjectList[i];
		ShaderObjectList[i] = NULL;
	}

	for(unsigned int i=0; i<ShaderProgramList.size(); i++) 
	{
		delete ShaderProgramList[i];
		ShaderProgramList[i] = NULL;
	}

	fclose(fp);
}

void GLSLManager::SetRenderTarget(void)
{
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	current_rt = NULL;
}

void GLSLManager::SetRenderTarget(RenderTexture* NewTarget)
{
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, NewTarget->GetFBOHandle());
	current_rt = NewTarget;
}

RenderTexture* GLSLManager::GetCurrentRenderTarget()
{
	return current_rt;
}

void GLSLManager::SetShader(void) 
{
	if(CurrentProgram == NullProgram)
		return;

	glUseProgram(0);
	CurrentProgram = NullProgram;
}
void GLSLManager::SetShader(const string &VSName, const string &GSName, const string &PSName) 
{
	//Find a program with a matched compiled shader
	for(unsigned int i=0; i<ShaderProgramList.size(); i++)
	{
		if( ShaderProgramList[i]->GetVS()->GetName() == VSName && 
			ShaderProgramList[i]->GetGS()->GetName() == GSName && 
			ShaderProgramList[i]->GetPS()->GetName() == PSName)
		{
			if(CurrentProgram == ShaderProgramList[i])
				return;

			glUseProgram(ShaderProgramList[i]->GetProgramHandle());
			CurrentProgram = ShaderProgramList[i];
			return;
		}
	}

	//No Program Matches, compile and form new program
	ShaderObject *VSObject = CompileShader(VSName);
	ShaderObject *GSObject = CompileShader(GSName);
	ShaderObject *PSObject = CompileShader(PSName);

	//No Match, create a new Program
	ShaderProgram* NewProgram = new ShaderProgram(VSObject, GSObject, PSObject, glCreateProgram());
	CurrentProgram = NewProgram;
	glUseProgram(NewProgram->GetProgramHandle());

	printProgramInfoLog(NewProgram->GetProgramHandle());

	ShaderProgramList.push_back(NewProgram);
}

ShaderObject* GLSLManager::CompileShader(const string &ShaderName)
{
	for(unsigned int i=0; i<ShaderObjectList.size(); i++)
	{
		if(ShaderObjectList[i]->GetName() == ShaderName)
			return ShaderObjectList[i];
	}

	//Find which shaders to compile
	SHADERTYPE ShaderType;
	if(ShaderName.rfind(".vert")!=string::npos) 
		ShaderType = VS;
	else if (ShaderName.rfind(".geom")!=string::npos)
		ShaderType = GS;
	else if (ShaderName.rfind(".frag")!=string::npos)
		ShaderType = PS;
	else {
		return NULL;
	}

	//Compile and set if not found
	ShaderObject* NewShader = new ShaderObject(ShaderName, ShaderType);

	//Temporarily change the current directory to look for a shader at a right place...
	GetCurrentDirectory(MAX_PATH, tmp_current_dir);
	SetCurrentDirectory(base_dir.c_str());

	NewShader->Compile();
	printShaderInfoLog(NewShader->GetShaderHandle());

	//Put back the current directory
	SetCurrentDirectory(tmp_current_dir);

	ShaderObjectList.push_back(NewShader);
	return NewShader;
}

void GLSLManager::SetTextureToIdx(unsigned int texID, GLuint hTexture)
{
	glActiveTexture(GL_TEXTURE0 + texID);
	glBindTexture(GL_TEXTURE_2D, hTexture);

	string texName = FormatToString("tex%d", texID);
	glUniform1i(glGetUniformLocation(CurrentProgram->GetProgramHandle(),texName.c_str()), texID);
}

void GLSLManager::SetTextureToIdx(unsigned int texID, GLuint hTexture, GLint min_mag_filter)
{
	glActiveTexture(GL_TEXTURE0 + texID);
	glBindTexture(GL_TEXTURE_2D, hTexture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, min_mag_filter);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min_mag_filter);

	string texName = FormatToString("tex%d", texID);
	glUniform1i(glGetUniformLocation(CurrentProgram->GetProgramHandle(),texName.c_str()), texID);
}

void GLSLManager::ResetTextures()
{
	for(unsigned int i=0; i<8; i++)
	{
		glActiveTexture(GL_TEXTURE0+i);
		glBindTexture(GL_TEXTURE_2D, 0);
	}
	glActiveTexture(GL_TEXTURE0);
}

void GLSLManager::printShaderInfoLog(GLuint obj)
{
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

	glGetShaderiv(obj, GL_INFO_LOG_LENGTH,&infologLength);

    if (infologLength > 1)
    {
        infoLog = (char *)malloc(infologLength);
        glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);

		fprintf(fp,"Shader Error Log:\n");
		fprintf(fp,"%s\n",infoLog);

        free(infoLog);
		AfxPostQuitMessage(1);
    }

}

void GLSLManager::printProgramInfoLog(GLuint obj)
{ 
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

	glGetProgramiv(obj, GL_INFO_LOG_LENGTH,&infologLength);

	
    if (infologLength > 1)
    {
        infoLog = (char *)malloc(infologLength);
        glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);

		fprintf(fp,"Program Error Log:\n");
		fprintf(fp,"%s\n",infoLog);

        free(infoLog);
		AfxPostQuitMessage(1);
    }

}


GLuint GLSLManager::GetCurrentProgram() {
	if(CurrentProgram == NullProgram) return 0;
	else return CurrentProgram->GetProgramHandle();
}
