#include "stdafx.h"
#include "IBFVRenderer.h"
#include <algorithm>
#include <time.h>
#include "cv.h"
#include "cvaux.h"
#include "StrokeMapManager.h"
#include "shader_flag.h"
#include "renderer_debug.h"
#include "VisView.h"
#include "ppm.h"


#if D_SAVESTAT_IBFV == 1
#include "debug_timing.h"
#else 
#include "debug_no_timing.h"
#endif

int d_ibfv_num_layers = 0;
int d_ibfv_num_steps = 0, d_ibfv_num_warps = 0, d_ibfv_num_composite = 0, d_ibfv_num_initimgcomps = 0;
int d_ibfv_num_refimgcomps, d_ibfv_num_cvblur_cvcopy = 0, d_ibfv_num_clears = 0, d_ibfv_num_listcomps = 0;
__int64 **d_ibfv_step = NULL, **d_ibfv_warp = NULL, **d_ibfv_composite = NULL, **d_ibfv_initimgcomps = NULL;
__int64 **d_ibfv_refimgcomp = NULL, **d_ibfv_cvblur_cvcopy = NULL, **d_ibfv_clear = NULL, **d_ibfv_listcomp = NULL;


typedef list<SimpleStroke*> seed_list;



/*
		ref:		0 ~ 255
		noise:		0 ~ 1.f
*/
inline void IBFV_sampleColor(unsigned char color[3], unsigned char ref[3], float noise[3], float amp) {
	color[0] = (unsigned char)::clampVal((ref[0] + noise[0] * amp), 0, 255);
	color[1] = (unsigned char)::clampVal((ref[1] + noise[1] * amp), 0, 255);
	color[2] = (unsigned char)::clampVal((ref[2] + noise[2] * amp), 0, 255);
}


/*
		ref:		0 ~ 255
		shift:		0 ~ 255.f
		noise:		0 ~ 1.f
*/
inline void IBFV_sampleColorJitter(unsigned char color[3], unsigned char ref[3], float shift[3], float noise[3], float amp) {
	/**
	color[0] = (unsigned char)::clampVal((ref[0] + noise[0] * amp), 0, 255);
	color[1] = (unsigned char)::clampVal((ref[1] + noise[1] * amp), 0, 255);
	color[2] = (unsigned char)::clampVal((ref[2] + noise[2] * amp), 0, 255);
	/**/

	/**/
	float hsv[3];
	RgbHsv(ref, hsv);

	// never jitter color when a color if fully-desaturated
	if(hsv[1] > 0 && hsv[2] > 0) {
		hsv[0] = hsv[0] + shift[0];
		hsv[1] = (float)::clampVal((hsv[1] + shift[1]), 0, 1.f);
		hsv[2] = (float)::clampVal((hsv[2] + shift[2]), 0, 1.f);
	}

	HsvRgb(hsv, color);
	color[0] = (unsigned char)::clampVal((color[0] + noise[0] * amp), 0, 255);
	color[1] = (unsigned char)::clampVal((color[1] + noise[1] * amp), 0, 255);
	color[2] = (unsigned char)::clampVal((color[2] + noise[2] * amp), 0, 255);
	/**/
}


// assume the triplet is in RGB order.
inline float calcDepth(unsigned char color[3]) {
	__int64 max_depth = 0x000002fa03ffffff;
	__int64 unscaled = ((__int64)(color[0]*color[0]+color[1]*color[1]+color[2]*color[2]) << 24) + ((__int64)color[1] << 16) + ((__int64)color[0] << 8) + (__int64)color[2];
	return unscaled / (float)max_depth;
}


IBFV_SimpleLayer::IBFV_SimpleLayer(int width, int height) {
	rendered_image = new unsigned char[4 * width * height];	
	initial_stroke_image = new unsigned char[4 * width * height];
	strokes.resize(0);
}


IBFV_SimpleLayer::~IBFV_SimpleLayer() {
	delete[] rendered_image;
	delete[] initial_stroke_image;

	while(strokes.empty() == false) {
		delete strokes.front();
		strokes.pop_front();
	}
}




const float IBFVRenderer::quad_vertex_coords[] = {
	-1.f, -1.f, 
	 1.f, -1.f, 
	 1.f,  1.f, 
	-1.f,  1.f, 
};

const float IBFVRenderer::quad_texture_coords[] = {
	0.0, 0.0,
	1.f, 0.0,
	1.f, 1.f, 
	0.0, 1.f, 
};


IBFVRenderer::IBFVRenderer(int width, int height, int num_layers) : ILayeredStrokePaintRenderer(width, height, num_layers) {
	initializeBuffer();
	makeSeedTemplate();

	use_gl_nearest_in_warp = false;
}


IBFVRenderer::~IBFVRenderer() {
	finalizeBuffer();
	cleanSeedTemplate();
}


void IBFVRenderer::initializeBuffer() {
	int num_elems;
	const float noise_gen_factor = 2.f / (float)RAND_MAX;

	srand(0);

	num_pixels = width * height;
	num_elems = 3 * num_pixels;
	color_noise = new float[num_elems];
	ref_image = new unsigned char[num_elems];
	layers = new IBFV_SimpleLayer*[max_layers];

	stroke_ref_img_ptr = NULL;

	for(int i = 0; i < num_elems; i++) {
		color_noise[i] = rand() * noise_gen_factor - 1.f;
	}


	for(int i = 0; i < max_layers; i++) {
		layers[i] = new IBFV_SimpleLayer(width, height);
	}


	for(int i = 0; i < 4; i++) {
		render_buffer[i] = new unsigned char[4 * num_pixels];
	}


	for(int i = 0; i < 2 ;i++) {
		aux[i] = new unsigned char[4 * num_pixels];
	}


	g_stroke_spacing = 0;
	g_stroke_falloff = 85;
	g_stroke_lifetime = 10;
	g_stroke_fadetime = 10;
	g_stroke_color = 0;
	current_layer = 0;
	bgfg_threshold = 28;
	render_start = 0;
	render_end = 0;
	QueryPerformanceFrequency((LARGE_INTEGER*)&timer_freq);

	_renderOnLayer = NULL;


	initial_stroke_image = new RenderTexture(width, height);
	offscreen = new RenderTexture(width, height);
	quad = new QuadVBO();


#if USE_GLSL_IBFV == 1
	render_buffer_RT[0] = new RenderTexture(width, height);
	render_buffer_RT[1] = new RenderTexture(width, height);
	render_buffer_RT[2] = new RenderTexture(width, height);
	render_buffer_RT[3] = new RenderTexture(width, height);
	tmp_target_RT = new RenderTexture(width, height);

	// preload shader program
	GLSLManager::SetShader("Basic.vert", "", "HsvComp.frag");
	GLSLManager::SetShader("Basic.vert", "", "RgbComp.frag");
	GLSLManager::SetShader("Disk.vert", "", "Disk_1.frag");
	GLSLManager::SetShader();


#else
	render_buffer_RT[0] = NULL;
	render_buffer_RT[1] = NULL;
	render_buffer_RT[2] = NULL;
	render_buffer_RT[3] = NULL;
	tmp_target_RT = NULL;

#endif

	// set up timing environment
	d_ibfv_num_steps = 2 + max_layers * 6;
	d_ibfv_num_refimgcomps = 3 * max_layers;
	d_ibfv_num_initimgcomps = 2 * max_layers;
	d_ibfv_num_cvblur_cvcopy = 3 * max_layers;

#if D_USE_DEPTH_IBFV == 0
	d_ibfv_num_listcomps = 4 * max_layers;
#else
	d_ibfv_num_listcomps = 3 * max_layers;
#endif

	d_ibfv_num_warps = (3 + 4) * max_layers;
	d_ibfv_num_composite = max_layers * 3;
	d_ibfv_num_clears = 2;
	D_INIT_TIMING(d_ibfv_step, d_ibfv_num_steps)
	D_INIT_TIMING(d_ibfv_refimgcomp, d_ibfv_num_refimgcomps)
	D_INIT_TIMING(d_ibfv_cvblur_cvcopy, d_ibfv_num_cvblur_cvcopy)
	D_INIT_TIMING(d_ibfv_initimgcomps, d_ibfv_num_initimgcomps)
	D_INIT_TIMING(d_ibfv_listcomp, d_ibfv_num_listcomps)
	D_INIT_TIMING(d_ibfv_warp, d_ibfv_num_warps)
	D_INIT_TIMING(d_ibfv_composite, d_ibfv_num_composite)
	D_INIT_TIMING(d_ibfv_clear, d_ibfv_num_clears)
}


void IBFVRenderer::finalizeBuffer() {
	delete[] color_noise;
	delete[] ref_image;

	for(int i = 0; i < max_layers; i++) {
		delete layers[i];
	}

	delete[] layers;


	for(int i = 0; i < 4; i++) {
		delete[] render_buffer[i];
	}


	for(int i = 0; i < 2; i++) {
		delete[] aux[i];
	}


	delete initial_stroke_image;
	delete offscreen;
	delete quad;
	delete render_buffer_RT[0];
	delete render_buffer_RT[1];
	delete render_buffer_RT[2];
	delete render_buffer_RT[3];
	delete tmp_target_RT;


	D_FIN_TIMING(d_ibfv_step)
	D_FIN_TIMING(d_ibfv_refimgcomp)
	D_FIN_TIMING(d_ibfv_cvblur_cvcopy)
	D_FIN_TIMING(d_ibfv_initimgcomps)
	D_FIN_TIMING(d_ibfv_listcomp)
	D_FIN_TIMING(d_ibfv_warp)
	D_FIN_TIMING(d_ibfv_composite)
	D_FIN_TIMING(d_ibfv_clear)
}


void IBFVRenderer::setParams(IPaintingParam *param) {
	this->param = *static_cast<IBFVParam*>(param);
}


bool IBFVRenderer::render(unsigned char *result) {
	int prev_seed_order = param.seed_order;
	RGBComparator::FUNC_TYPE prev_rgb_functype = StrokeComparator::rgb_comp.getEvalFunc();
	RenderTexture *prev_rt;


	if(result == NULL) return false;

	painted_image = result;
	render_start = 0;
	render_end = 0;


	prev_rt = GLSLManager::GetCurrentRenderTarget();

	// set for this rendering
	GLSLManager::SetRenderTarget(offscreen);


	// post-processing setting.
	if(render_mode == 1 && param.embossing_type == 0) {
		param.seed_order = 1;
		StrokeComparator::rgb_comp.setEvalFunc(RGBComparator::MAX);
	}


	// select a renderer
	if(param.image_composition.image_composition_method == 0 && param.image_composition.copy_back == false) {
		_renderOnLayer = &IBFVRenderer::renderPaintingOnLayer;

		selectActiveShaderAndParams();
	}
	else {
		_renderOnLayer = &IBFVRenderer::renderPaintingOnLayerDebug;
	}


	// select which image to use to look up stroke color
	stroke_ref_img_ptr = param.stroke_ref_img ? param.stroke_ref_img : ref_image;


	// render
	QueryPerformanceCounter((LARGE_INTEGER*)&render_start);

	createLayeredWarpedImage();

	QueryPerformanceCounter((LARGE_INTEGER*)&render_end);


	// disconnect
	painted_image = NULL;
	_renderOnLayer = NULL;


	if(render_mode == 1 && param.embossing_type == 0) {
		param.seed_order = prev_seed_order;
		StrokeComparator::rgb_comp.setEvalFunc(prev_rgb_functype);
	}


	// restore
	if(prev_rt != NULL) {
		GLSLManager::SetRenderTarget(prev_rt);
	}
	else {
		GLSLManager::SetRenderTarget();
	}



#if D_SAVESTAT_IBFV == 1
	int d_step_idx = 0;
	int d_initimgcomp_idx = 0;
	int d_listcomp_idx = 0;
	int d_refimgcmp_idx = 0;
	int d_cvblur_copy_idx = 0;
	int d_warp_idx = 0;
	int d_composite_idx = 0;
	int d_clear_idx = 0;
	string path(VisView::getFullPath());
	path.append("\\IBFV_stat.txt");
	FILE *d_fp = fopen(path.c_str(), "a");


	if(render_mode == 0) {
		fprintf(d_fp, "Image rendering:                                         \tTime(sec)\n");
	}
	else {
		fprintf(d_fp, "Heightmap rendering:                                     \tTime(sec)\n");
	}


	fprintf(d_fp, "Initialize (clear seeds in unused layers):               \t%.3f\n\n", D_TIME(d_ibfv_step, d_step_idx, timer_freq)); d_step_idx++;

	for(int i = 0; i < d_ibfv_num_layers; i++) {
		fprintf(d_fp, "Layer %d\n", i);

		if(render_mode == 0) {
			fprintf(d_fp, "  Make ref. image:                                       \t\n", D_TIME(d_ibfv_step, d_step_idx, timer_freq)); d_step_idx++;
			fprintf(d_fp, "    Initialize:                                          \t%.3f\n", D_TIME(d_ibfv_refimgcomp, d_refimgcmp_idx, timer_freq)); d_refimgcmp_idx++;
			fprintf(d_fp, "    Blur:                                                \t\n", D_TIME(d_ibfv_refimgcomp, d_refimgcmp_idx, timer_freq)); d_refimgcmp_idx++;
			fprintf(d_fp, "      Setup:                                             \t%.3f\n", D_TIME(d_ibfv_cvblur_cvcopy, d_cvblur_copy_idx, timer_freq)); d_cvblur_copy_idx++;
			fprintf(d_fp, "      cvBlur():                                          \t%.3f\n", D_TIME(d_ibfv_cvblur_cvcopy, d_cvblur_copy_idx, timer_freq)); d_cvblur_copy_idx++;
			fprintf(d_fp, "      cvCopy():                                          \t%.3f\n", D_TIME(d_ibfv_cvblur_cvcopy, d_cvblur_copy_idx, timer_freq)); d_cvblur_copy_idx++;
			fprintf(d_fp, "    Copy & finalize:                                     \t%.3f\n", D_TIME(d_ibfv_refimgcomp, d_refimgcmp_idx, timer_freq)); d_refimgcmp_idx++;

			fprintf(d_fp, "  Compute a seed list:                                   \t\n", D_TIME(d_ibfv_step, d_step_idx, timer_freq)); d_step_idx++;
			fprintf(d_fp, "    Initialize:                                          \t%.3f\n", D_TIME(d_ibfv_listcomp, d_listcomp_idx, timer_freq)); d_listcomp_idx++;
			fprintf(d_fp, "    Generate seeds:                                      \t%.3f\n", D_TIME(d_ibfv_listcomp, d_listcomp_idx, timer_freq)); d_listcomp_idx++;
#if D_USE_DEPTH_IBFV == 0
			fprintf(d_fp, "    Sort:                                                \t%.3f\n", D_TIME(d_ibfv_listcomp, d_listcomp_idx, timer_freq)); d_listcomp_idx++;
#endif
			fprintf(d_fp, "    Assign depth:                                        \t%.3f\n", D_TIME(d_ibfv_listcomp, d_listcomp_idx, timer_freq)); d_listcomp_idx++;
		}
		else {
			d_step_idx++;
			d_step_idx++;
		}

		fprintf(d_fp, "  Make Initial image:                                    \t\n", D_TIME(d_ibfv_step, d_step_idx, timer_freq)); d_step_idx++;
		fprintf(d_fp, "    Draw:                                                \t%.3f\n", D_TIME(d_ibfv_initimgcomps, d_initimgcomp_idx, timer_freq)); d_initimgcomp_idx++;
		fprintf(d_fp, "    Save:                                                \t%.3f\n", D_TIME(d_ibfv_initimgcomps, d_initimgcomp_idx, timer_freq)); d_initimgcomp_idx++;

		fprintf(d_fp, "  Render:                                                \t\n", D_TIME(d_ibfv_step, d_step_idx, timer_freq)); d_step_idx++;
		fprintf(d_fp, "    Initialize:                                          \t%.3f\n", D_TIME(d_ibfv_warp, d_warp_idx, timer_freq)); d_warp_idx++;
		fprintf(d_fp, "    Image1:                                              \t%.3f\n", D_TIME(d_ibfv_warp, d_warp_idx, timer_freq)); d_warp_idx++;
		fprintf(d_fp, "    Image2:                                              \t%.3f\n", D_TIME(d_ibfv_warp, d_warp_idx, timer_freq)); d_warp_idx++;
		fprintf(d_fp, "    Image3:                                              \t%.3f\n", D_TIME(d_ibfv_warp, d_warp_idx, timer_freq)); d_warp_idx++;
		fprintf(d_fp, "    Image4:                                              \t%.3f\n", D_TIME(d_ibfv_warp, d_warp_idx, timer_freq)); d_warp_idx++;
		fprintf(d_fp, "    Composite:                                           \t\n", D_TIME(d_ibfv_warp, d_warp_idx, timer_freq)); d_warp_idx++;
		
		fprintf(d_fp, "      1st:                                               \t%.3f\n", D_TIME(d_ibfv_composite, d_composite_idx, timer_freq)); d_composite_idx++;
		fprintf(d_fp, "      2nd:                                               \t%.3f\n", D_TIME(d_ibfv_composite, d_composite_idx, timer_freq)); d_composite_idx++;
		fprintf(d_fp, "      3rd:                                               \t%.3f\n", D_TIME(d_ibfv_composite, d_composite_idx, timer_freq)); d_composite_idx++;

		fprintf(d_fp, "    Finalize:                                            \t%.3f\n", D_TIME(d_ibfv_warp, d_warp_idx, timer_freq)); d_warp_idx++;

		fprintf(d_fp, "  Composite with canvas:                                 \t%.3f\n", D_TIME(d_ibfv_step, d_step_idx, timer_freq)); d_step_idx++;
		fprintf(d_fp, "  Scale edge field:                                      \t%.3f\n", D_TIME(d_ibfv_step, d_step_idx, timer_freq)); d_step_idx++;
		fprintf(d_fp, "\n");
	}

	d_step_idx = 1 + max_layers * 6;
	fprintf(d_fp, "Cleanup:                                                 \t%\n", D_TIME(d_ibfv_step, d_step_idx, timer_freq)); d_step_idx++;
	fprintf(d_fp, "  Release memory for extra seeds:                        \t%.3f\n", D_TIME(d_ibfv_clear, d_clear_idx, timer_freq)); d_clear_idx++;
	fprintf(d_fp, "  Restore edge field on mesh:                            \t%.3f\n", D_TIME(d_ibfv_clear, d_clear_idx, timer_freq)); d_clear_idx++;
	fprintf(d_fp, "---------------------------------------------------------------\n");
	fprintf(d_fp, "Render Time:                                             \t%.3f\n", (double)(d_ibfv_step[1][d_step_idx - 1]-d_ibfv_step[0][0]) /(double)timer_freq);

	if(render_mode == 0) {
		fprintf(d_fp, "Number of strokes: [");

		for(int i = 0; i < max_layers; i++) {
			fprintf(d_fp, "%d", getNumStrokes(i));

			if(i < max_layers - 1) fprintf(d_fp, ", ");
		}

		fprintf(d_fp, "]\n");
	}

	fprintf(d_fp, "\n\n");
	fclose(d_fp);

#endif


	return true;
}


void IBFVRenderer::selectActiveShaderAndParams() {
#if USE_GLSL_IBFV == 1
	switch(param.seed_order) {
		case 0:
		case 1:
		default:
			active_shader_name = "RgbComp.frag";
			
			switch(StrokeComparator::rgb_comp.getEvalFunc()) {
				case RGBComparator::MIN:
				default:
					shader_func_eval_params[0] = 0;
					break;

				case RGBComparator::MAX:
					shader_func_eval_params[0] = 1;
					break;
			}

			shader_func_eval_params[1] = 0;
			shader_func_eval_params[2] = 0;
			shader_func_eval_params[3] = 0;
			break;

		case 2:
			active_shader_name = "HsvComp.frag";

			switch(StrokeComparator::hsv_comp.getEvalFunc_H()) {
				case HSVComparator::MIN:
				default:
					shader_func_eval_params[0] = 0;
					break;

				case HSVComparator::MAX:
					shader_func_eval_params[0] = 1;
					break;

				case HSVComparator::WARM1:
					shader_func_eval_params[0] = 2;
					break;

				case HSVComparator::COOL1:
					shader_func_eval_params[0] = 3;
					break;

				case HSVComparator::WARM2:
					shader_func_eval_params[0] = 4;
					break;

				case HSVComparator::COOL2:
					shader_func_eval_params[0] = 5;
					break;
			}

			switch(StrokeComparator::hsv_comp.getEvalFunc_S()) {
				case HSVComparator::MIN:
				default:
					shader_func_eval_params[1] = 0;
					break;

				case HSVComparator::MAX:
					shader_func_eval_params[1] = 1;
					break;
			}

			switch(StrokeComparator::hsv_comp.getEvalFunc_V()) {
				case HSVComparator::MIN:
				default:
					shader_func_eval_params[2] = 0;
					break;

				case HSVComparator::MAX:
					shader_func_eval_params[2] = 1;
					break;
			}

			switch(StrokeComparator::hsv_comp.getEvalOrder()) {
				case HSVComparator::HSV:
				default:
					shader_func_eval_params[3] = 0;
					break;

				case HSVComparator::SVH:
					shader_func_eval_params[3] = 1;
					break;

				case HSVComparator::VHS:
					shader_func_eval_params[3] = 2;
					break;

				case HSVComparator::VSH:
					shader_func_eval_params[3] = 3;
					break;

				case HSVComparator::HVS:
					shader_func_eval_params[3] = 4;
					break;

				case HSVComparator::SHV:
					shader_func_eval_params[3] = 5;
					break;
			}

			break;
	}
#endif
}


void IBFVRenderer::passUniformVariables() {
#if USE_GLSL_IBFV == 1
	float scaled_bgfg_threshold = bgfg_threshold / 255.f;

	switch(param.seed_order) {
		case 0:
		case 1:
		default:
			glUniform1f(glGetUniformLocation(GLSLManager::GetCurrentProgram(), "bgfg_threshold"), scaled_bgfg_threshold);
			glUniform1i(glGetUniformLocation(GLSLManager::GetCurrentProgram(), "func"), shader_func_eval_params[0]);
			break;

		case 2:
			glUniform1f(glGetUniformLocation(GLSLManager::GetCurrentProgram(), "bgfg_threshold"), scaled_bgfg_threshold);
			glUniform1f(glGetUniformLocation(GLSLManager::GetCurrentProgram(), "hue_breakpoint"), StrokeComparator::hsv_comp.getHueBreakPoint());
			glUniform1i(glGetUniformLocation(GLSLManager::GetCurrentProgram(), "h_func"), shader_func_eval_params[0]);
			glUniform1i(glGetUniformLocation(GLSLManager::GetCurrentProgram(), "s_func"), shader_func_eval_params[1]);
			glUniform1i(glGetUniformLocation(GLSLManager::GetCurrentProgram(), "v_func"), shader_func_eval_params[2]);
			glUniform1i(glGetUniformLocation(GLSLManager::GetCurrentProgram(), "eval"), shader_func_eval_params[3]);
			break;
	}
#endif
}


bool IBFVRenderer::resize(int width, int height) {
	if(0 <= width || 0 <= height) return false;

	finalizeBuffer();

	this->width = width;
	this->height = height;

	initializeBuffer();

	return true;
}


bool IBFVRenderer::setNumLayers(int num_layers) {
	if(0 <= num_layers) return false;

	return false;
}


void IBFVRenderer::pushSeeds(int layer, const float *opt_flow, int width, int height) {
	int x, y, idx;
	int h_1 = this->height - 1;
	seed_list *ls;
	seed_list::iterator it, itr_end;
	SimpleStroke *stroke;

	if(opt_flow == NULL || width <= 0 || height <= 0) return;


	if(0 <= layer && layer < max_layers) {
		ls = &(layers[layer]->strokes);
		it = ls->begin();
		itr_end = ls->end();

		for(; it != itr_end; it++) {
			stroke = reinterpret_cast<SimpleStroke*>(*it);
			x = (int)stroke->x;
			y = (int)(h_1 - stroke->y);

			if(x < 0 || x >= this->width || y < 0 || y >= this->height) {	// should be within rendering window
				stroke->valid = false;
			}
			else if(x < width && y < height) {								// shift if it is within defined optical flow
				idx = 2 * (x + width * y);

				stroke->x = (float)(int)opt_flow[idx];
				stroke->y = (float)(int)(h_1 - opt_flow[idx + 1]);

				if(stroke->x < 0 || stroke->x >= this->width || stroke->y < 0 || stroke->y >= this->height) {
					stroke->valid = false;
				}
			}
		}
	}
}


size_t IBFVRenderer::getNumStrokes(int layer) {
	if(0 <= layer && layer < max_layers) {
		return this->layers[layer]->strokes.size();
	}
	else {
		return 0;
	}
}


double IBFVRenderer::getRenderingTime() {
	return (double)(render_end - render_start) / (double)timer_freq;
}


void IBFVRenderer::clearSeeds(int layer_idx) {
	if(0 <= layer_idx && layer_idx < max_layers) {
		s_stroke_factory.dump(layers[layer_idx]->strokes);
	}
}


const unsigned char* IBFVRenderer::getLayerImage(int layer) {
	if(0 <= layer && layer < max_layers) {
		return (const unsigned char*)layers[layer]->rendered_image;
	}
	else {
		return NULL;
	}
}


const unsigned char* IBFVRenderer::getInitialImage(int layer) {
	if(0 <= layer && layer < max_layers) {
		return (const unsigned char*)layers[layer]->initial_stroke_image;
	}
	else {
		return NULL;
	}
}


void IBFVRenderer::makeSeedTemplate() {
	float delta, theta;
	int idx;

	dl_unit_circle = 0;
	dl_circle_tex = 0;
	dl_draw_seed = 0;
	stroke_shape_handle = 0;
	
	for(int i = 0; i < num_height_patterns; i++) {
		stroke_tex_handle[i] = 0;
	}

	circular_stroke_coords = NULL;
	circular_stroke_tex_coords = NULL;
	stroke_tex_scaled_coords = NULL;
	stroke_jitter_patterns = NULL;


	circular_stroke_coords = new float[2 * num_circular_stroke_coords];
	circular_stroke_tex_coords = new float[2 * num_circular_stroke_coords];
	delta = 2.f * (float)M_PI / num_circular_stroke_coords;

	for(int i = 0; i < num_circular_stroke_coords; i++) {
		theta = i * delta;
		idx = 2 * i;

		circular_stroke_coords[idx] = (float)cos(theta);
		circular_stroke_coords[idx + 1] = (float)sin(theta);

		circular_stroke_tex_coords[idx] = (circular_stroke_coords[idx] + 1) * .5f;
		circular_stroke_tex_coords[idx + 1] = (circular_stroke_coords[idx + 1] + 1) * .5f;
	}


#if USE_GLSL_IBFV == 1
	dl_unit_circle = glGenLists(1);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);
	glNewList(dl_unit_circle, GL_COMPILE);
	glTexCoordPointer(2, GL_FLOAT, 0, circular_stroke_tex_coords);
	glVertexPointer(2, GL_FLOAT, 0, circular_stroke_coords);
	glDrawArrays(GL_POLYGON, 0, num_circular_stroke_coords);
	glEndList();
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	dl_draw_seed = glGenLists(1);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);
	glNewList(dl_draw_seed, GL_COMPILE);

#if D_SEEDING_IBFV == 0
	glStencilFunc(GL_EQUAL, 0, 3);
	glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
	glBlendFunc(GL_ONE, GL_ZERO);
	glCallList(dl_unit_circle);
	glStencilFunc(GL_EQUAL, 3, 3);
	glStencilOp(GL_REPLACE, GL_KEEP, GL_KEEP);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
#endif

	glCallList(dl_unit_circle);
	glEndList();
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

#else
	stroke_tex_scaled_coords = new float[2 * num_circular_stroke_coords]; 

#endif


	stroke_tex_handle[0] = StrokeMapManager::makeStrokeHeightmapTexture("./ibfv_stroke_height0.bmp");
	stroke_tex_handle[1] = StrokeMapManager::makeStrokeHeightmapTexture("./ibfv_stroke_height1.bmp");
	stroke_tex_handle[2] = StrokeMapManager::makeStrokeHeightmapTexture("./ibfv_stroke_height2.bmp");
}


void IBFVRenderer::cleanSeedTemplate() {
	// clean up seed primitive
	glDeleteLists(dl_unit_circle, 1);
	glDeleteLists(dl_circle_tex, 1);
	glDeleteLists(dl_draw_seed, 1);

	delete[] circular_stroke_coords;
	delete[] circular_stroke_tex_coords;
	delete[] stroke_tex_scaled_coords;

	glDeleteTextures(1, &(stroke_shape_handle));

	for(int i = 0; i < num_height_patterns; i++) {
		glDeleteTextures(1, &(stroke_tex_handle[i]));
	}

	if(stroke_jitter_patterns != NULL) {
		for(int i = 0; i < num_height_patterns; i++) {
			delete[] stroke_jitter_patterns[i];
		}

		delete[] stroke_jitter_patterns;
	}
}


void IBFVRenderer::createLayeredWarpedImage() {
	int idx1, idx2;
	int num_layers;
	int num_elems = 4 * num_pixels;

	const float decay_factor = .5;
	const float alpha_scale_factor = 1.f / 255.f;
	const float falloff = g_stroke_falloff / 100.f;
	const float blend_factor_constant = alpha_scale_factor / (1.f - falloff);

	float stroke_blend_factor, omsb;
	float alpha_max, alpha;
	float separataion_threshold;

	unsigned char *read_target = NULL;

	PaintingStyleNode *psn = param.style->getStyleNode();
	Polyhedron *poly = param.mesh_info.mesh;


	D_SET_COUNTER_VAR(d_cnt_idx, 0)


	// initialization
	D_TICK(d_ibfv_step, d_cnt_idx)

	num_layers = param.style->getMaxLayers();

	for(int i = num_layers; i < max_layers; i++) {
		clearSeeds(i);
	}

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);


	// global stroke parameter setting...
	if(param.style->getNumRegions() == 1) {
		g_stroke_spacing = psn[0].preset.stroke_gridsize;
	}
	else {
		g_stroke_spacing = param.style->getStrokeSpacingForMultiStyle();
	}

	g_stroke_color = 0;

	D_TOCK(d_ibfv_step, d_cnt_idx++)
	d_ibfv_num_layers = num_layers;


	idx1 = 0;

	for(int i = 0; i < num_pixels; i++) {
		painted_image[idx1++] = 255;
		painted_image[idx1++] = 255;
		painted_image[idx1++] = 255;
	}


	// core part... i assume that the edge field is given as a scaled
	for(int i = 0; i < num_layers; i++) {
		current_layer = i;

		// I assume that reference image is already available for height-map rendering... (i.e. the color-painting is already generated)
		D_TICK(d_ibfv_step, d_cnt_idx)

		if(render_mode == 0) {
			makeRefImage();
		}

		D_TOCK(d_ibfv_step, d_cnt_idx++)


		// make a stroke list (skipped for post-processing)
		D_TICK(d_ibfv_step, d_cnt_idx)
		if(render_mode == 0) {
			if(seed_gen_mode == 0) {
				regenerateStrokeList();
			}
			else {
				updateStrokeList();
			}
		}
		D_TOCK(d_ibfv_step, d_cnt_idx++)


		// make an initial image
		D_TICK(d_ibfv_step, d_cnt_idx)
		makeInitialImage();
		D_TOCK(d_ibfv_step, d_cnt_idx++)


		// render
		D_TICK(d_ibfv_step, d_cnt_idx)
		(this->*_renderOnLayer)();
		D_TOCK(d_ibfv_step, d_cnt_idx++)



		// composite
		D_TICK(d_ibfv_step, d_cnt_idx)

		if(i == 0) {
			idx1 = 0;
			idx2 = 0;
			read_target = layers[current_layer]->rendered_image;

			for(int j = 0; j < num_pixels; j++) {
				painted_image[idx1++] = read_target[idx2++];
				painted_image[idx1++] = read_target[idx2++];
				painted_image[idx1++] = read_target[idx2++];
				idx2++;
			}
		}
		else {
			idx1 = 0;
			idx2 = 0;
			read_target = layers[current_layer]->rendered_image;

			for(int j = 0; j < num_pixels; j++, idx1 += 3, idx2 += 4) {
				alpha = read_target[idx2 + 3];
				alpha_max = psn[j].preset.stroke_initial_opacity;
				separataion_threshold = alpha_max * falloff;

				if(alpha >= separataion_threshold) {
					if(alpha > alpha_max) alpha = alpha_max;
					stroke_blend_factor = (alpha - separataion_threshold) * blend_factor_constant;

					omsb = 1.f - stroke_blend_factor;

					painted_image[idx1] = (unsigned char)(stroke_blend_factor * read_target[idx2] + omsb * painted_image[idx1]);
					painted_image[idx1 + 1] = (unsigned char)(stroke_blend_factor * read_target[idx2 + 1] + omsb * painted_image[idx1 + 1]);
					painted_image[idx1 + 2] = (unsigned char)(stroke_blend_factor * read_target[idx2 + 2] + omsb * painted_image[idx1 + 2]);
				}
			}
		}
		
		D_TOCK(d_ibfv_step, d_cnt_idx++)
		
		
		// update a map used to locate positions to seed new strokes in the next layer.
		D_TICK(d_ibfv_step, d_cnt_idx)

		if(i < num_layers - 1) {
			g_stroke_spacing *= decay_factor;
			scaleTexCoords(decay_factor);
		}

		D_TOCK(d_ibfv_step, d_cnt_idx++)
	}


	D_UPDATE_COUNTER_VAR(d_cnt_idx, 1 + 6 * max_layers)
	D_TICK(d_ibfv_step, d_cnt_idx)


	// throw extra seeds
	D_TICK(d_ibfv_clear, 0)
	s_stroke_factory.hold();
	D_TOCK(d_ibfv_clear, 0)

	// restore edge field
	D_TICK(d_ibfv_clear, 1)
	restoreTexCoords();
	D_TOCK(d_ibfv_clear, 1)


	D_TOCK(d_ibfv_step, d_cnt_idx++)

}


void IBFVRenderer::scaleTexCoords(float scale) {
	Polyhedron *poly = param.mesh_info.mesh;

	for(int i = 0; i < poly->nverts; i++) {
		poly->vlist[i]->stepsize *= scale;
	}

	poly->calc_new_texture_coors();						// we need to recompute this since we change stepsize
	copyTexCoordsList();
}


void IBFVRenderer::restoreTexCoords() {
	int px, py;
	int w_1 = width - 1;
	int h_1 = height - 1;
	int num_verts;
	PaintingStyleNode *psn;
	Polyhedron *poly;
	Vertex *v;

	psn = param.style->getStyleNode();
	poly = param.mesh_info.mesh;
	num_verts = poly->nverts;

	for(int i = 0; i < num_verts; i++) {
		v = poly->vlist[i];
		px = (int)::clampVal((v->x + .5) * width, 0, w_1);
		py = (int)::clampVal((v->y + .5) * height, 0, h_1);
		v->stepsize = (float)(psn[px + width * py].preset.displacement);
	}

	poly->calc_new_texture_coors();
	copyTexCoordsList();
}


void IBFVRenderer::copyTexCoordsList() {
	int idx;
	int num_visible_triangles;
	int *tri_idx_ptr;
	float *tex_coords_array;
	double *v_st;
	Triangle *t;
	Polyhedron *poly;

	num_visible_triangles = param.mesh_info.num_visible_triangles;
	poly = param.mesh_info.mesh;
	tri_idx_ptr = param.mesh_info.visible_tri_idx;

	for(int i = 0; i < 8; i++) {
		idx = 0;
		tex_coords_array = param.mesh_info.visible_tri_vertex_tex_coords[i];

		for(int j = 0; j < num_visible_triangles; j++) {
			t = poly->tlist[tri_idx_ptr[j]];

			v_st = t->verts[0]->new_ts[i];
			tex_coords_array[idx++] = (float)v_st[0] + 0.5f;
			tex_coords_array[idx++] = (float)v_st[1] + 0.5f;

			v_st = t->verts[1]->new_ts[i];
			tex_coords_array[idx++] = (float)v_st[0] + 0.5f;
			tex_coords_array[idx++] = (float)v_st[1] + 0.5f;

			v_st = t->verts[2]->new_ts[i];
			tex_coords_array[idx++] = (float)v_st[0] + 0.5f;
			tex_coords_array[idx++] = (float)v_st[1] + 0.5f;
		}
	}
}


void IBFVRenderer::makeRefImage() {
	int idx1, idx2;
	int preset_idx = 0;
	int kernel_size = 1;
	int num_elems = 3 * num_pixels;
	
	IplImage *ipl_img = NULL;
	IplImage *ipl_ref = NULL;
	IplImage *ipl_mask = NULL;
	char *ipl_img_origin;
	
	PaintingStyleNode *psn = param.style->getStyleNode();
	int num_styles = (int)param.style->getNumRegions();

	map<int, PaintingRegion*>::iterator it, it_end;
	PaintingRegion *pr =NULL;
	int region_id;



	D_SET_COUNTER_VAR(d_cnt_idx, 3 * current_layer)
	D_SET_COUNTER_VAR(d_cnt_idx2, 0)


	D_TICK(d_ibfv_refimgcomp, d_cnt_idx)

	ipl_img = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U, 3);
	ipl_ref = cvCloneImage(ipl_img);
	ipl_mask = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U, 1);
	ipl_img_origin = ipl_img->imageDataOrigin;

	it_end = param.style->getEnd();

	D_TOCK(d_ibfv_refimgcomp, d_cnt_idx++)


	D_TICK(d_ibfv_refimgcomp, d_cnt_idx)

	// create a reference image by smoothing input. this image will be used in later
	// let sigma be computed from kernel_size
	for(it = param.style->getBegin(); it != it_end; it++) {

		D_UPDATE_COUNTER_VAR(d_cnt_idx2, 3 * current_layer)		// time is measured on only one region...
		D_TICK(d_ibfv_cvblur_cvcopy, d_cnt_idx2)

		pr = it->second;
		region_id = pr->getID();

		for(int j = 0; j < num_elems; j++) {
			ipl_img_origin[j] = (char)param.source_img[j];
		}

		kernel_size = 1;

		for(int j = 0; j < height; j++) {
			idx1 = width * j;
			idx2 = width * (height - 1 - j);

			for(int k = 0; k < width; k++) {
				if(region_id == psn[idx2].region_id) {
					ipl_mask->imageDataOrigin[idx1++] = 1;
					kernel_size = (int)(psn[idx2++].preset.stroke_size);
				}
				else {
					ipl_mask->imageDataOrigin[idx1++] = 0;
					idx2++;
				}
			}
		}

		// make it odd-size
		if(((kernel_size >> 1) << 1) == kernel_size) kernel_size++;

		D_TOCK(d_ibfv_cvblur_cvcopy, d_cnt_idx2++)

		D_TICK(d_ibfv_cvblur_cvcopy, d_cnt_idx2)
		cvSmooth(ipl_img, ipl_img, CV_GAUSSIAN, kernel_size, 0);		// this will be slow if kernel size is too large
		D_TOCK(d_ibfv_cvblur_cvcopy, d_cnt_idx2++)

		D_TICK(d_ibfv_cvblur_cvcopy, d_cnt_idx2)
		cvCopy(ipl_img, ipl_ref, ipl_mask);
		D_TOCK(d_ibfv_cvblur_cvcopy, d_cnt_idx2++)
	}

	D_TOCK(d_ibfv_refimgcomp, d_cnt_idx++)


	D_TICK(d_ibfv_refimgcomp, d_cnt_idx)

	// upside-down
	ipl_img_origin = ipl_ref->imageDataOrigin;

	for(int i = 0; i < height; i++) {
		idx1 = 3 * width * i;
		idx2 = 3 * width * (height - 1 - i);

		for(int j = 0; j < width; j++) {
			ref_image[idx1++] = (unsigned char)(ipl_img_origin[idx2++]);
			ref_image[idx1++] = (unsigned char)(ipl_img_origin[idx2++]);
			ref_image[idx1++] = (unsigned char)(ipl_img_origin[idx2++]);
		}
	}

	cvReleaseImage(&ipl_img);
	cvReleaseImage(&ipl_ref);
	cvReleaseImage(&ipl_mask);

	D_TOCK(d_ibfv_refimgcomp, d_cnt_idx++)
}


void IBFVRenderer::makeInitialImage() {

	D_SET_COUNTER_VAR(d_cnt_idx, 2 * current_layer)


	D_TICK(d_ibfv_initimgcomps, d_cnt_idx)
	seedStrokes();
	D_TOCK(d_ibfv_initimgcomps, d_cnt_idx++)

	D_TICK(d_ibfv_initimgcomps, d_cnt_idx)
	glReadPixels(0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, layers[current_layer]->initial_stroke_image);
	D_TOCK(d_ibfv_initimgcomps, d_cnt_idx++)


	/**
	if(render_mode == 1 && param.embossing_type == 0) {
		int n_elems = width * height;
		unsigned char *ptr = layers[current_layer]->initial_stroke_image;
		unsigned char *bg = new unsigned char[3 * n_elems];
		for(int i = 0; i < n_elems; i++) { bg[3*i] = ptr[4*i+0];bg[3*i+1] = ptr[4*i+1];bg[3*i+2] = ptr[4*i+2];}
		My_SSS_SavePPM(width, height, bg, "init_height.ppm");
		delete[] bg;
	}
	/**/


	/**
	if(current_layer == 0 && render_mode == 1) {
		int n_elems = width * height;
		unsigned char *ptr = layers[current_layer]->initial_stroke_image;
		unsigned char *bg = new unsigned char[3 * n_elems];
		for(int i = 0; i < n_elems; i++) { bg[3*i] = ptr[4*i+0];bg[3*i+1] = ptr[4*i+1];bg[3*i+2] = ptr[4*i+2];}
		My_SSS_SavePPM(width, height, bg, "init.ppm");
		delete[] bg;
	}
	/**/
}


void IBFVRenderer::regenerateStrokeList() {
	/* For now, assume that (width == height)... */
	int xi, yi;
	int w_1 = width - 1;
	int h_1 = height - 1;
	int psn_idx, idx_3;
	int push_id;
	int num_grids;
	int stroke_half_lifetme = g_stroke_lifetime + g_stroke_fadetime;
	float offset, center, delta;
	float half_rand = RAND_MAX * .5f;
	float decay_factor = 1.f;

	SimpleStroke *stroke = NULL;
	PaintingPreset *pp = NULL;
	vector< SimpleStroke* > seed_order;
	vector< SimpleStroke* >::iterator order_itr, order_end;

	PaintingStyleNode *psn = param.style->getStyleNode();
	FieldSampleImage *fsi = param.edge_pixel;
	FieldSample *fs = NULL;



#if D_USE_DEPTH_IBFV == 0
	D_SET_COUNTER_VAR(d_cnt_idx, 4 * current_layer)
#else
	D_SET_COUNTER_VAR(d_cnt_idx, 3 * current_layer)
#endif

	D_TICK(d_ibfv_listcomp,  d_cnt_idx)

	clearSeeds(current_layer);

	for(int i = 0; i < current_layer; i++) {
		decay_factor *= .5f;
	}


	// skip if either the stroke size or grid size is too small
	if(g_stroke_spacing < 1.f) return;

	num_grids = (int)ceil((double)width / g_stroke_spacing);
	delta = g_stroke_spacing;
	offset = (num_grids * delta - width) * .5f;
	center = delta * .5f - offset;

	srand(0);

	D_TOCK(d_ibfv_listcomp,  d_cnt_idx++)

		
	D_TICK(d_ibfv_listcomp,  d_cnt_idx)

	if(current_layer == 0) {
		const float jitter_scale_factor = delta / (100.f * RAND_MAX);
		float jitter_scale;

		for(int i = 0; i < num_grids; i++) {
			for(int j = 0; j < num_grids; j++) {
				xi = (int)::clampVal(center + j * delta, 0, w_1);
				yi = (int)::clampVal(center + i * delta, 0, h_1);
				psn_idx = xi + width * yi;
				//idx_3 = 3 * psn_idx;

				pp = &(psn[psn_idx].preset);
				//fs = &(fsi->fld[xi + width * (h_1 - yi)]);


				jitter_scale = pp->stroke_pos_jitter * jitter_scale_factor;

				xi = (int)::clampVal(xi + (int)(jitter_scale * (rand() - half_rand)), 0, w_1);
				yi = (int)::clampVal(yi + (int)(jitter_scale * (rand() - half_rand)), 0, h_1);
				push_id = rand();


				psn_idx = xi + width * yi;
				idx_3 = 3 * psn_idx;

				pp = &(psn[psn_idx].preset);
				fs = &(fsi->fld[xi + width * (h_1 - yi)]);


				stroke = s_stroke_factory.make(0);
				stroke->x = (float)xi;
				stroke->y = (float)yi;
				stroke->size = pp->stroke_size * decay_factor;
				stroke->lifetime = stroke_half_lifetme;

				stroke->stroke_jitter_shift[0] = color_noise[idx_3];
				stroke->stroke_jitter_shift[1] = color_noise[idx_3 + 1];
				stroke->stroke_jitter_shift[2] = color_noise[idx_3 + 2];

				stroke->jitter_id = push_id % 3;
				stroke->depth = push_id;

				//IBFV_sampleColorJitter(stroke->color, &(ref_image[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
				IBFV_sampleColorJitter(stroke->color, &(stroke_ref_img_ptr[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
				stroke->angle = (float)atan2(fs->vec_fld.entry[1], fs->vec_fld.entry[0]);

				seed_order.push_back(stroke);
			}
		}
	}
	else {
		int cnt;
		int tmp_psn_idx;
		int grid_size = (int)delta;
		int num_kernel_elems;
		int *xs = NULL, *ys = NULL;
		int kx, ky, kx0, ky0;
		float *color_diff = new float[num_pixels];
		float area_error;

		// make odd-size
		if(((grid_size >> 1) << 1) == grid_size) grid_size++;

		num_kernel_elems = grid_size * grid_size;
		xs = new int[num_kernel_elems];
		ys = new int[num_kernel_elems];

		// compute color difference between painted image and reference image
		computeColorDifference(painted_image, ref_image, width, height, color_diff);

		//if(current_layer == 1)saveFloatImg(width, height, 1, color_diff, _T("floatImgIBFV.ppm"));

		for(int i = 0; i < num_grids; i++) {
			for(int j = 0; j < num_grids; j++) {
				xi = (int)::clampVal(center + j * delta, 0, w_1);
				yi = (int)::clampVal(center + i * delta, 0, h_1);
				psn_idx = xi + width * yi;

				push_id = rand();

				if(current_layer >= psn[psn_idx].preset.num_layers) continue;

				area_error = 0;
				cnt = 0;

				findRegion(width, height, xi, yi, grid_size, xs, ys, false);

				for(int m = 0; m < num_kernel_elems; m++) {
					kx = xs[m];
					ky = ys[m];

					if(0 <= kx && kx < width && 0 <= ky && ky < height) {
						area_error += color_diff[kx + width * ky];
						cnt++;
					}
				}

				if(cnt > 0) {
					area_error /= cnt;

					if(psn[psn_idx].preset.area_error < area_error) {
						psn_idx = -1;

						// find a pixel with the largest error
						for(int m = 0; m < num_kernel_elems; m++) {
							kx = xs[m];
							ky = ys[m];

							if(0 <= kx && kx < width && 0 <= ky && ky < height) {
								tmp_psn_idx = kx + width * ky;

								if(psn_idx < 0 || color_diff[tmp_psn_idx] > color_diff[psn_idx]) {
									psn_idx = tmp_psn_idx;
									kx0 = kx;
									ky0 = ky;
									//break;
								}
							}
						}

						if(psn_idx >= 0) {
							idx_3 = 3 * psn_idx;
							pp = &(psn[psn_idx].preset);
							fs = &(fsi->fld[kx0 + width * (h_1 - ky0)]);

							stroke = s_stroke_factory.make(0);
							stroke->x = (float)kx0;
							stroke->y = (float)ky0;
							stroke->size = pp->stroke_size * decay_factor;
							stroke->lifetime = stroke_half_lifetme;

							stroke->stroke_jitter_shift[0] = color_noise[idx_3];
							stroke->stroke_jitter_shift[1] = color_noise[idx_3 + 1];
							stroke->stroke_jitter_shift[2] = color_noise[idx_3 + 2];

							stroke->jitter_id = push_id % 3;
							stroke->depth = push_id;

							//IBFV_sampleColorJitter(stroke->color, &(ref_image[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
							IBFV_sampleColorJitter(stroke->color, &(stroke_ref_img_ptr[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
							stroke->angle = (float)atan2(fs->vec_fld.entry[1], fs->vec_fld.entry[0]);

							seed_order.push_back(stroke);
						}
					}
				}
			}
		}

		delete[] xs;
		delete[] ys;
		delete[] color_diff;
	}

	D_TOCK(d_ibfv_listcomp,  d_cnt_idx++)



#if D_USE_DEPTH_IBFV == 0

	D_TICK(d_ibfv_listcomp,  d_cnt_idx)

	switch(param.seed_order) {
		case 0:
		default:
			std::sort(seed_order.begin(), seed_order.end(), StrokeComparator::compareStrokeDepth);
			break;

		case 1:
			std::sort(seed_order.begin(), seed_order.end(), StrokeComparator::compareStrokeRGB);
			break;

		case 2:
			// extra step to compute hsv color...
			order_itr = seed_order.begin();
			order_end = seed_order.end();

			for(; order_itr != order_end; order_itr++) {
				RgbHsv((*order_itr)->color, (*order_itr)->hsv);
			}

			std::sort(seed_order.begin(), seed_order.end(), StrokeComparator::compareStrokeHSV);
			break;
	}

	D_TOCK(d_ibfv_listcomp,  d_cnt_idx++)


	D_TICK(d_ibfv_listcomp,  d_cnt_idx)

	// create strokes
	for(size_t i = 0; i < seed_order.size(); i++) {
		layers[current_layer]->strokes.push_back(seed_order[i]);
	}

	D_TOCK(d_ibfv_listcomp,  d_cnt_idx++)

#else

	D_TICK(d_ibfv_listcomp,  d_cnt_idx)

	switch(param.seed_order) {
		case 0:
		default:
			for(size_t i = 0; i < seed_order.size(); i++) {
				seed_order[i]->z = seed_order[i]->depth / (float)RAND_MAX;
				layers[current_layer]->strokes.push_back(seed_order[i]);
			}

			break;

		case 1:
			for(size_t i = 0; i < seed_order.size(); i++) {
				seed_order[i]->z = calcDepth(seed_order[i]->color);
				layers[current_layer]->strokes.push_back(seed_order[i]);
			}

			break;
	}

	D_TOCK(d_ibfv_listcomp,  d_cnt_idx++)

#endif

}


void IBFVRenderer::updateStrokeList() {
	/* For now, assume that (width == height)... */
	int xi, yi;
	int w_1 = width - 1;
	int h_1 = height - 1;
	int psn_idx, idx_3;
	int push_id;
	int stroke_half_lifetime = g_stroke_fadetime + g_stroke_lifetime;
	int stroke_total_lifetime = 2 * g_stroke_fadetime + g_stroke_lifetime;
	int num_grids;
	int removal_cnt[2] = {0, 0};
	float offset, delta;
	SimpleStroke *stroke;
	PaintingPreset *pp = NULL;
	vector< SimpleStroke* > seed_order;
	vector< SimpleStroke* >::iterator order_itr, order_end;


	float decay_factor = 1.f;
	PaintingStyleNode *psn = param.style->getStyleNode();


	seed_list *ls;
	seed_list::iterator it, itr_end;

	unsigned char *stroke_map = NULL;

	int num_subgrids;
	int hs;
	int cover;
	int tx, ty;
	bool flag;


	FieldSampleImage *fsi = param.edge_pixel;
	FieldSample *fs = NULL;


	for(int i = 0; i < current_layer; i++) {
		decay_factor *= .5f;
	}

	float min_stroke_size = param.style->findMinStrokeSize(param.time);
	hs = ((int)min_stroke_size >> 1) + 1;
	cover = 2 * hs;



	//----------------------------------------------------------------------------------------------------------//
	//		Compute stroke map
	//----------------------------------------------------------------------------------------------------------//


	// remove all invalid strokes that are pushed outside of canvas
	ls = &(layers[current_layer]->strokes);
	it = ls->begin();
	itr_end = ls->end();

	if(current_layer == 0) {
		for(; it != itr_end; ) {
			if((*it)->valid == false) {
				s_stroke_factory.dump(*it);
				it = ls->erase(it);
				removal_cnt[0]++;
			}
			else {
				it++;
			}
		}
	}
	else {
		for(; it != itr_end; ) {
			(*it)->lifetime--;

			if((*it)->lifetime <= 0 || (*it)->valid == false) {
				s_stroke_factory.dump(*it);
				it = ls->erase(it);
				removal_cnt[0]++;
			}
			else {
				it++;
			}
		}
	}


	// return here
	if(min_stroke_size <= 0) return;


	// merge crowded seeds
	//...how to deal with the transparency?
	int bin_size, sqrt_bin_size;
	float d, ood;
	int containment_threshold = 1;
	vector<SimpleStroke*> *s_bin;
	vector<SimpleStroke*>::iterator itv;

	if(current_layer == 0) {
		d = (float)floor(min_stroke_size / (2 * 1.42));
		
		if(d <= 0) d = 1;
	}
	else{
		d = g_stroke_spacing;
	}

	ood = 1 / d;

	sqrt_bin_size = (int)ceil((float)width * ood);
	bin_size = sqrt_bin_size * sqrt_bin_size;
	s_bin = new vector<SimpleStroke*>[bin_size];


	containment_threshold = (int)ceil(d / ((min_stroke_size < 1) ? 1.f : min_stroke_size));
	containment_threshold *= containment_threshold;

	// We want to hold more stroke seeds that have transparency than those with 0 opacity..
	// Here comes a factor to scale the containment threshold...
	//
	// Factor is defined by 255 / (stroke_alpha)
	//
	// For simplicity, assume stroke opacity doesn't go below 128... 
	//
	containment_threshold *= 2;	
	itr_end = ls->end();

	for(it = ls->begin(); it != itr_end; it++) {
		xi = (int)::clampVal((*it)->x * ood, 0, sqrt_bin_size - 1);
		yi = (int)::clampVal((*it)->y * ood, 0, sqrt_bin_size - 1);

		s_bin[xi + sqrt_bin_size * yi].push_back(*it);
	}


	for(int i = 0; i < bin_size; i++) {
		if(s_bin[i].size() > (size_t)containment_threshold) {

#if D_USE_DEPTH_IBFV == 1
			std::sort(s_bin[i].begin(), s_bin[i].end(), ::compareStrokeDepth);
#endif 
			for(size_t i = containment_threshold; i < s_bin[i].size(); i++) {
				s_bin[i].at(i)->valid = false;
			}
		}
	}


	itr_end = ls->end();

	for(it = ls->begin(); it != itr_end; ) {
		if((*it)->valid == false) {
			s_stroke_factory.dump(*it);
			it = ls->erase(it);
			removal_cnt[1]++;
		}
		else {
			it++;
		}
	}


	// recompute colors and angles (orientation)
	itr_end = ls->end();

	for(it = ls->begin(); it != itr_end; it++) {
		psn_idx = ((int)(*it)->x) + width * ((int)(*it)->y);
		idx_3 = 3 * psn_idx;
		pp = &(psn[psn_idx].preset);
		//IBFV_sampleColorJitter((*it)->color, &(ref_image[idx_3]), pp->stroke_color_shift, (*it)->stroke_jitter_shift, pp->jitter_amplitude);
		IBFV_sampleColorJitter((*it)->color, &(stroke_ref_img_ptr[idx_3]), pp->stroke_color_shift, (*it)->stroke_jitter_shift, pp->jitter_amplitude);

		fs = &(fsi->fld[((int)(*it)->x) + width * (h_1 - ((int)(*it)->y))]);

		if(cos((*it)->angle) * fs->vec_fld.entry[0] + sin((*it)->angle) * fs->vec_fld.entry[1] < 0) {
			(*it)->angle = (float)atan2(-1 * fs->vec_fld.entry[1], -1 * fs->vec_fld.entry[0]);
		}
		else {
			(*it)->angle = (float)atan2(fs->vec_fld.entry[1], fs->vec_fld.entry[0]);
		}
	}


	delete[] s_bin;



	if(g_stroke_spacing < 1.f) return;

	stroke_map = new unsigned char[3 * num_pixels];

	// seed black strokes with no color jitter
	g_stroke_color = 1;
	seedStrokes();
	g_stroke_color = 0;


	glReadPixels(0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, layers[current_layer]->initial_stroke_image);

	(this->*_renderOnLayer)();

	// read from pixels_total (not from frame buffer)
	for(int i = 0; i < num_pixels; i++) {
		stroke_map[i] = layers[current_layer]->rendered_image[4 * i];
	}


	/**
	if(current_layer == 1) {
		int idx = 0;
		unsigned char *bg = new unsigned char[3 * num_pixels];
		for(int i = 0; i < num_pixels; i++, idx += 3) {
			bg[idx] = stroke_map[i];
			bg[idx + 1] = stroke_map[i];
			bg[idx + 2] = stroke_map[i];
		}
		My_SSS_SavePPM(width, height, bg, _T("spline_strokemap_ibfv.ppm"));
		delete[] bg;
	}
	/**/



	//----------------------------------------------------------------------------------------------------------//
	//		Find where to put new seeds
	//----------------------------------------------------------------------------------------------------------//
	num_grids = (int)ceil((double)width / g_stroke_spacing);
	offset = (num_grids *  g_stroke_spacing - width) * .5f;

	srand(0);

	if(current_layer == 0) {
		float corner;
		delta = (float)floor(min_stroke_size / (2 * 1.42));

		if(delta <= 0) delta = 1;

		num_subgrids = (int)ceil(width / (double)delta);

		// it's not exactly a center at all... rather, it is a corner position.
		corner = (width - num_subgrids * delta) * .5f;

		for(int i = 0; i < num_subgrids; i++) {
			for(int j = 0; j < num_subgrids; j++) {
				xi = (int)(corner + j * delta);
				yi = (int)(corner + i * delta);

				push_id = rand();
				flag = false;

				for(int m = 0; m < delta && flag == false; m++) {
					ty = yi + m;

					if(0 <= ty && ty < height) {
						for(int n = 0; n < delta && flag == false; n++) {
							tx = xi + n;

							if(0 <= tx && tx < width) {
								if(stroke_map[tx + width * ty] > bgfg_threshold) {
									flag = true;

									// seed a stroke at first empty pixel
									xi = tx;
									yi = ty;
									psn_idx = xi + width * yi;
									idx_3 = 3 * psn_idx;
									pp = &(psn[psn_idx].preset);
									fs = &(fsi->fld[xi + width * (h_1 - yi)]);

									stroke = s_stroke_factory.make(0);
									stroke->x = (float)xi;
									stroke->y = (float)yi;
									stroke->size = pp->stroke_size * decay_factor;
									stroke->lifetime = stroke_half_lifetime;

									stroke->stroke_jitter_shift[0] = color_noise[idx_3];
									stroke->stroke_jitter_shift[1] = color_noise[idx_3 + 1];
									stroke->stroke_jitter_shift[2] = color_noise[idx_3 + 2];

									stroke->jitter_id = push_id % 3;
									stroke->depth = push_id;

									//IBFV_sampleColorJitter(stroke->color, &(ref_image[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
									IBFV_sampleColorJitter(stroke->color, &(stroke_ref_img_ptr[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
									stroke->angle = (float)atan2(fs->vec_fld.entry[1], fs->vec_fld.entry[0]);

									seed_order.push_back(stroke);
								}
							}
						}
					}
				}
			}
		}
	}
	else {
		int cnt;
		int tmp_psn_idx;
		int grid_size;
		int num_kernel_elems;
		int *xs = NULL, *ys = NULL;
		int kx, ky, kx0, ky0;
		float *color_diff = new float[num_pixels];
		float area_error;
		float center;


		delta = g_stroke_spacing;
		num_subgrids = num_grids;		// won't be used...
		center = delta * .5f - offset;
		grid_size = (int)delta;


		// make odd-size
		if(((grid_size >> 1) << 1) == grid_size) grid_size++;

		num_kernel_elems = grid_size * grid_size;
		xs = new int[num_kernel_elems];
		ys = new int[num_kernel_elems];

		// compute color difference between painted image and reference image
		computeColorDifference(painted_image, ref_image, width, height, color_diff);

		for(int i = 0; i < num_grids; i++) {
			for(int j = 0; j < num_grids; j++) {
				xi = (int)::clampVal(center + j * delta, 0, w_1);
				yi = (int)::clampVal(center + i * delta, 0, h_1);
				psn_idx = xi + width * yi;

				push_id = rand();

				area_error = 0;
				cnt = 0;

				findRegion(width, height, xi, yi, grid_size, xs, ys, false);

				for(int m = 0; m < num_kernel_elems; m++) {
					kx = xs[m];
					ky = ys[m];

					if(0 <= kx && kx < width && 0 <= ky && ky < height) {
						area_error += color_diff[kx + width * ky];
						cnt++;
					}
				}

				if(cnt > 0) {
					area_error /= cnt;

					if(psn[psn_idx].preset.area_error < area_error) {
						psn_idx = -1;

						// find a pixel with the largest error
						for(int m = 0; m < num_kernel_elems; m++) {
							kx = xs[m];
							ky = ys[m];

							if(0 <= kx && kx < width && 0 <= ky && ky < height) {
								tmp_psn_idx = kx + width * ky;

								if(psn_idx < 0 || color_diff[tmp_psn_idx] > color_diff[psn_idx]) {
									psn_idx = tmp_psn_idx;
									kx0 = kx;
									ky0 = ky;
								}
							}
						}

						if(psn_idx >= 0 && stroke_map[psn_idx] > bgfg_threshold && current_layer < psn[psn_idx].preset.num_layers) {
							idx_3 = 3 * psn_idx;
							pp = &(psn[psn_idx].preset);
							fs = &(fsi->fld[kx0 + width * (h_1 - ky0)]);

							stroke = s_stroke_factory.make(0);
							stroke->x = (float)kx0;
							stroke->y = (float)ky0;
							stroke->size = pp->stroke_size * decay_factor;
							//stroke->lifetime = stroke_half_lifetime;
							stroke->lifetime = stroke_total_lifetime;


							stroke->stroke_jitter_shift[0] = color_noise[idx_3];
							stroke->stroke_jitter_shift[1] = color_noise[idx_3 + 1];
							stroke->stroke_jitter_shift[2] = color_noise[idx_3 + 2];

							stroke->jitter_id = push_id % 3;
							stroke->depth = push_id;

							//IBFV_sampleColorJitter(stroke->color, &(ref_image[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
							IBFV_sampleColorJitter(stroke->color, &(stroke_ref_img_ptr[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
							stroke->angle = (float)atan2(fs->vec_fld.entry[1], fs->vec_fld.entry[0]);

							seed_order.push_back(stroke);
						}
					}
				}
			}
		}

		delete[] xs;
		delete[] ys;
		delete[] color_diff;
	}


	ls = &(layers[current_layer]->strokes);
	itr_end = ls->end();


	// copy over for sort a whole set of seeds...
	if(param.partial_ordering == false) {
		for(it = ls->begin(); it != itr_end; it++) {
			seed_order.push_back(*it);
		}

		ls->clear();
	}


	switch(param.seed_order) {
		case 0:
		default:
			std::sort(seed_order.begin(), seed_order.end(), StrokeComparator::compareStrokeDepth);
			break;

		case 1:
			std::sort(seed_order.begin(), seed_order.end(), StrokeComparator::compareStrokeRGB);
			break;

		case 2:
			// extra step to compute hsv color...
			order_itr = seed_order.begin();
			order_end = seed_order.end();

			for(; order_itr != order_end; order_itr++) {
				RgbHsv((*order_itr)->color, (*order_itr)->hsv);
			}

			std::sort(seed_order.begin(), seed_order.end(), StrokeComparator::compareStrokeHSV);
			break;
	}



	for(size_t i = 0; i < seed_order.size(); i++) {
		ls->push_back(seed_order[i]);
	}

	delete[] stroke_map;
}


void IBFVRenderer::setUpMultiTexturePainting() {
#if USE_GLSL_IBFV == 0
	glActiveTexture(GL_TEXTURE0);
	glDisable(GL_TEXTURE_2D);
#endif
}


void IBFVRenderer::cleanUpMultiTexturePainting() {
#if USE_GLSL_IBFV == 0
	glActiveTexture(GL_TEXTURE0);
	glDisable(GL_TEXTURE_2D);
#endif
}


void IBFVRenderer::setUpMultiTexturePostProcess() {
#if USE_GLSL_IBFV == 0
	// stage 0: scale z-value of height map (scaling is applied as tex. env. color)
	glActiveTexture(GL_TEXTURE0);
	glEnable(GL_TEXTURE_2D);
	glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &(prev_tex_env_mode[0]));
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
	glTexEnvi(GL_TEXTURE_ENV, GL_SRC0_RGB, GL_TEXTURE);
	glTexEnvi(GL_TEXTURE_ENV, GL_SRC1_RGB, GL_CONSTANT);
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);


	// stage 1: subtract shift from stroke color (primary color), add opacity
	glActiveTexture(GL_TEXTURE1);
	glEnable(GL_TEXTURE_2D);
	glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &(prev_tex_env_mode[1]));
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_SUBTRACT);
	glTexEnvi(GL_TEXTURE_ENV, GL_SRC0_RGB, GL_PRIMARY_COLOR);
	glTexEnvi(GL_TEXTURE_ENV, GL_SRC1_RGB, GL_PREVIOUS);
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
	glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA, GL_REPLACE);
	glTexEnvi(GL_TEXTURE_ENV, GL_SRC0_ALPHA, GL_PRIMARY_COLOR);
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA, GL_SRC_COLOR);

	glActiveTexture(GL_TEXTURE0);

#endif
}


void IBFVRenderer::cleanUpMultiTexturePostProcess() {
#if USE_GLSL_IBFV == 0
	glActiveTexture(GL_TEXTURE1);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, prev_tex_env_mode[1]);
	glDisable(GL_TEXTURE_2D);
	glActiveTexture(GL_TEXTURE0);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, prev_tex_env_mode[0]);
	glDisable(GL_TEXTURE_2D);

#endif
}


void IBFVRenderer::setBackgroundColor() {
	float hsv[3], rgb[3];
	float hc;

	if(render_mode == 0 || param.embossing_type == 1) {
		/**
		switch(param.seed_order) {
			case 0:
			case 1:
			default:
				switch(StrokeComparator::rgb_comp.getEvalFunc()) {
					case RGBComparator::MIN:
					default:
						glClearColor(1, 1, 1, 0);
						break;

					case RGBComparator::MAX:
						glClearColor(0, 0, 0, 0);
						break;
				}

				break;

			case 2:
				hc = StrokeComparator::hsv_comp.getHueBreakPoint();

				switch(StrokeComparator::hsv_comp.getEvalFunc_H()) {
					case HSVComparator::MIN:
						hsv[0] = hc + 360.f;
						break;

					case HSVComparator::MAX:
						hsv[0] = hc;
						break;

					case HSVComparator::WARM1:
					case HSVComparator::WARM2:
						hsv[0] = hc + 180.f;

						break;

					case HSVComparator::COOL1:
					case HSVComparator::COOL2:
						hsv[0] = hc;
						break;
				}

				switch(StrokeComparator::hsv_comp.getEvalFunc_S()) {
					case HSVComparator::MIN:
					default:
						hsv[1] = 1.f;
						break;

					case HSVComparator::MAX:
						hsv[1] = 0;
						break;
				}

				switch(StrokeComparator::hsv_comp.getEvalFunc_V()) {
					case HSVComparator::MIN:
					default:
						hsv[2] = 1.f;
						break;

					case HSVComparator::MAX:
						hsv[2] = 0;
						break;
				}

				HsvRgb(hsv, rgb);
				glClearColor(rgb[0], rgb[1], rgb[2], 0);

				break;
		}
		/**/
		glClearColor(1.f, 1.f, 1.f, 0);
	}
	else {
		glClearColor(0, 0, 0, 0);
	}
}


void IBFVRenderer::seedStrokes() {
	int ch = 4;
	int w_1 = width - 1;
	int h_1 = height - 1;
	int idx1, idx2;
	float x, y;
	unsigned char rgb_ub[3];
	float r, g, b, a;
	float theta;

	seed_list *ls;
	seed_list::reverse_iterator it;
	seed_list::reverse_iterator itr_end;
	FieldSample *fs = NULL;

	const float degree = 180 / (float)M_PI;

	float br;
	float radius;
	float stroke_color_jitter, stroke_color_jitter_freq;

	float scaled_stroke_falloff = g_stroke_falloff / 100.f;
	float frac = 1 / 255.f;
	float oow = 1.f / width;
	float ooh = 1.f / height;
	float sf = oow;			// scale factor
	int idx_3;

	float decay_factor = 1.f;
	PaintingStyleNode *fsn = param.style->getStyleNode();
	PaintingPreset *pp;

	GLuint program_id, jitter_scale_loc, jitter_freq_loc;


	int num_strokes = 0, cnt = 0;
	float offset, dh;
	float min_offset = .2f;
	float max_offset = .4f;



#if D_SEEDING_IBFV == 0
	glEnable(GL_STENCIL_TEST);
	glClear(GL_STENCIL_BUFFER_BIT);

#endif


	setBackgroundColor();

	for(int i = 0; i < max_layers; i++) {
		num_strokes += layers[i]->strokes.size();

		if(i < current_layer) cnt += layers[i]->strokes.size();
	}

	dh = (max_offset - min_offset) / (num_strokes + 1);



	glEnable(GL_BLEND);


#if D_USE_DEPTH_IBFV == 0
	glClear(GL_COLOR_BUFFER_BIT);

#else
	glEnable(GL_DEPTH_TEST);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

#endif
	


#if D_SEEDING_IBFV == 1
	glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_ONE, GL_ZERO);
#endif


	ls = &(layers[current_layer]->strokes);
	it = ls->rbegin();
	itr_end = ls->rend();


	for(int i = 0; i < current_layer; i++) {
		decay_factor *= .5f;
	}


#if USE_GLSL_IBFV == 1
	GLSLManager::SetShader("Disk.vert", "", "Disk_1.frag");

	program_id = GLSLManager::GetCurrentProgram();
	jitter_scale_loc = glGetUniformLocation(program_id, "jitter_scale");
	jitter_freq_loc = glGetUniformLocation(program_id, "jitter_freq");

	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);

#else

	switch(render_mode) {
		case 0:
			setUpMultiTexturePainting();
			break;

		case 1:
			setUpMultiTexturePostProcess();
			break;
	}


	// enable vertex array and 3 tex. coord. arrays
	glClientActiveTexture(GL_TEXTURE0);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE1);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	glEnableClientState(GL_VERTEX_ARRAY);


#endif



	// seed strokes in reverse order
	for(; it != itr_end; it++) {
		// Assume (x, y) is within [0,0]x[w_1,h_1]
		x = (*it)->x;
		y = (*it)->y;

		idx1 = (int)x + width * (int)y;
		idx2 = (int)x + width * (h_1 - (int)y);

		idx_3 = idx1 * 3;

		x = x * oow - .5f;
		y = y * ooh - .5f;
		fs = &(param.edge_pixel->fld[idx2]);
		pp = &(fsn[idx1].preset);


		if(render_mode == 1) {
			stroke_color_jitter = pp->post_process_val3;
			stroke_color_jitter_freq = pp->post_process_val4;
		}
		else {
			stroke_color_jitter = 0;
			stroke_color_jitter_freq = 0;
		}


		//theta = (float)atan2(fs->vec_fld.entry[1], fs->vec_fld.entry[0]);
		theta = (*it)->angle;

		// convert from pixel unit to gl-space
		radius = pp->stroke_size * decay_factor * .5f;
		br = radius * sf;

		switch(g_stroke_color) {
			case 0:
				rgb_ub[0] = (*it)->color[0];
				rgb_ub[1] = (*it)->color[1];
				rgb_ub[2] = (*it)->color[2];

				break;

			case 1:
				rgb_ub[0] = 0;
				rgb_ub[1] = 0;
				rgb_ub[2] = 0;

				break;
		}


		if(render_mode == 0 || param.embossing_type == 1) {
			r = rgb_ub[0] * frac;
			g = rgb_ub[1] * frac;
			b = rgb_ub[2] * frac;
			a = calcAlpha(pp->stroke_initial_opacity * frac, scaled_stroke_falloff, g_stroke_lifetime, g_stroke_fadetime, (*it)->lifetime);
		}
		else {
			offset = min_offset + (cnt++) * dh;
			r = offset;
			b = offset;
			g = offset;
			//a = 1.f;
			a = calcAlpha(pp->stroke_initial_opacity * frac, scaled_stroke_falloff, g_stroke_lifetime, g_stroke_fadetime, (*it)->lifetime);
		}


		glPushMatrix();

#if D_USE_DEPTH_IBFV == 0
		glTranslatef(x, y, 0);
#else
		glTranslatef(x, y, -(*it)->z);
#endif


		glRotatef(degree * theta, 0, 0, 1);
		glScalef(br, br, 1);


#if USE_GLSL_IBFV == 1
		// larger value for stroke_color_jitter_freq gets slower...
		//if(g_stroke_color == 0 && stroke_color_jitter > 0 && stroke_color_jitter_freq > .1) {
		if(g_stroke_color == 0 && stroke_color_jitter > 0 && stroke_color_jitter_freq > 0) {
			glUniform1f(jitter_scale_loc, stroke_color_jitter);
			glUniform1f(jitter_freq_loc, stroke_color_jitter_freq);
		}
		else {
			glUniform1f(jitter_scale_loc, 0);
			glUniform1f(jitter_freq_loc, 0);
		}

		glColor4f(r, g, b, a);
		GLSLManager::SetTextureToIdx(0, stroke_tex_handle[(*it)->jitter_id]);


#if D_SEEDING_IBFV == 0
		glCallList(dl_draw_seed);
#else
		glCallList(dl_unit_circle);
#endif



#else

		if(g_stroke_color != 0 || stroke_color_jitter <= 0 || stroke_color_jitter_freq <= 0) {
			stroke_color_jitter = 0;
			stroke_color_jitter_freq = 0;
		}

		drawSeed(stroke_color_jitter, stroke_color_jitter_freq, stroke_tex_handle[(*it)->jitter_id], r, g, b, a);

#endif

		glPopMatrix();
	}




#if USE_GLSL_IBFV == 1
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	GLSLManager::ResetTextures();
	GLSLManager::SetShader();

#else

	// disable
	glDisableClientState(GL_VERTEX_ARRAY);

	glClientActiveTexture(GL_TEXTURE1);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE0);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);


	switch(render_mode) {
		case 0:
			cleanUpMultiTexturePainting();
			break;

		case 1:
			cleanUpMultiTexturePostProcess();
			break;
	}

#endif


#if D_SEEDING_IBFV == 0
	glDisable(GL_STENCIL_TEST);
#endif

	glDisable(GL_BLEND);

#if D_USE_DEPTH_IBFV == 1
	glDisable(GL_DEPTH_TEST);
#endif

}


void IBFVRenderer::drawSeed(float jitter_scale, float jitter_freq, GLuint seed_tex_handle, float r, float g, float b, float a) {
#if USE_GLSL_IBFV == 0

	int idx = 0;
	float tex_env_color[4];
	float frac = 1.f / 255.f;


	glColor4f(r, g, b, a);

	tex_env_color[0] = jitter_scale * frac;
	tex_env_color[1] = tex_env_color[0];
	tex_env_color[2] = tex_env_color[0];
	tex_env_color[3] = 1.f;

	for(int i = 0; i < num_circular_stroke_coords; i++) {
		stroke_tex_scaled_coords[idx] = circular_stroke_tex_coords[idx] * jitter_freq; idx++;
		stroke_tex_scaled_coords[idx] = circular_stroke_tex_coords[idx] * jitter_freq; idx++;
	}

	glActiveTexture(GL_TEXTURE0);
	glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, tex_env_color);
	glBindTexture(GL_TEXTURE_2D, seed_tex_handle);

	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, 0);

	glClientActiveTexture(GL_TEXTURE0);
	glTexCoordPointer(2, GL_FLOAT, 0, stroke_tex_scaled_coords);

	glClientActiveTexture(GL_TEXTURE1);
	glTexCoordPointer(2, GL_FLOAT, 0, circular_stroke_tex_coords);

#if D_SEEDING_IBFV == 0
	glStencilFunc(GL_EQUAL, 0, 3);
	glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
	glBlendFunc(GL_ONE, GL_ZERO);

	glVertexPointer(2, GL_FLOAT, 0, circular_stroke_coords);
	glDrawArrays(GL_POLYGON, 0, num_circular_stroke_coords);

	glStencilFunc(GL_EQUAL, 3, 3);
	glStencilOp(GL_REPLACE, GL_KEEP, GL_KEEP);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

#endif

	glVertexPointer(2, GL_FLOAT, 0, circular_stroke_coords);
	glDrawArrays(GL_POLYGON, 0, num_circular_stroke_coords);

	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, 0);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, 0);

#endif
}


#if USE_GLSL_IBFV == 1

void IBFVRenderer::renderPaintingOnLayer() {
	GLuint var_loc = -1;
	int num_verts = 3 * param.mesh_info.num_visible_triangles;
	int *coords = param.image_composition.coords;

	GLint viewport[4];
	unsigned char *render_target = layers[current_layer]->rendered_image;
	unsigned char *tmp_arr2[4];

	D_SET_COUNTER_VAR(d_cnt_idx, current_layer * 7)

	// initialize
	D_TICK(d_ibfv_warp, d_cnt_idx)

	for(int i = 0; i < 4; i++) {
		tmp_arr2[i] = render_buffer[i];
	}


	glEnable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);

	glClearColor(1, 1, 1, 0);
	glClear(GL_COLOR_BUFFER_BIT);

	glGetIntegerv(GL_VIEWPORT, viewport);
	glViewport(0, 0, width, height);

	// initialize
	GLSLManager::SetShader();
	GLSLManager::ResetTextures();
	glBindTexture(GL_TEXTURE_2D, initial_stroke_image->GetTextureHandle());
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, layers[current_layer]->initial_stroke_image);


	for(int g = 0; g < 4; g++) {
		GLSLManager::SetRenderTarget(render_buffer_RT[g]);	
		quad->Draw();
	}


	GLSLManager::ResetTextures();


	D_TOCK(d_ibfv_warp, d_cnt_idx++)



	// Warp individual 4 images.
	for(int g = 0; g < 4; g++) {
		D_TICK(d_ibfv_warp, d_cnt_idx)

		for(int h = 0; h < param.num_iterations; h++) {
			GLSLManager::SetShader();
			GLSLManager::SetRenderTarget(tmp_target_RT);

			glBindTexture(GL_TEXTURE_2D, render_buffer_RT[g]->GetTextureHandle());

			if(use_gl_nearest_in_warp) {
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			}

			glEnableClientState(GL_VERTEX_ARRAY);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glVertexPointer(2, GL_FLOAT, 0, param.mesh_info.visible_tri_vertex_coords);
			glTexCoordPointer(2, GL_FLOAT, 0, param.mesh_info.visible_tri_vertex_tex_coords[coords[g]]);
			glDrawArrays(GL_TRIANGLES, 0, num_verts);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			glDisableClientState(GL_VERTEX_ARRAY);


			if(use_gl_nearest_in_warp) {
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			}

			GLSLManager::SetShader("Basic.vert", "", active_shader_name);
			//GLSLManager::SetTextureToIdx(0, initial_stroke_image->GetTextureHandle(), GL_NEAREST);
			//GLSLManager::SetTextureToIdx(1, tmp_target_RT->GetTextureHandle(), GL_NEAREST);
			GLSLManager::SetTextureToIdx(0, initial_stroke_image->GetTextureHandle());
			GLSLManager::SetTextureToIdx(1, tmp_target_RT->GetTextureHandle());
			passUniformVariables();

			GLSLManager::SetRenderTarget(render_buffer_RT[g]);

			quad->Draw();

			GLSLManager::ResetTextures();
		}

		glReadPixels(0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, tmp_arr2[g]);

		D_TOCK(d_ibfv_warp, d_cnt_idx++)
	}


	/**
	unsigned char *bg = new unsigned char[3 * width * height];
	int d_idx0 = 0, d_idx1 = 0;
	stringstream tmp_filename;
	for(int i = 0; i < 4; i++) {
		d_idx0 = 0;
		d_idx1 = 0;
		for(int j = 0; j < num_pixels; j++) {
			bg[d_idx0++] = tmp_arr2[i][d_idx1++];
			bg[d_idx0++] = tmp_arr2[i][d_idx1++];
			bg[d_idx0++] = tmp_arr2[i][d_idx1++];
			d_idx1++;
		}

		tmp_filename.str("");

		switch(i) {
			case 0:
				tmp_filename << "warp_(-x).ppm";
				break;

			case 1:
				tmp_filename << "warp_(-y).ppm";
				break;

			case 2:
				tmp_filename << "warp_(+x).ppm";
				break;

			case 3:
				tmp_filename << "warp_(+y).ppm";
				break;
		}

		My_SSS_SavePPM(width, height, bg, tmp_filename.str().c_str());
	}
	delete[] bg;
	/**/


	// composite
	D_TICK(d_ibfv_warp, d_cnt_idx)

	overwriteColor3(tmp_arr2, param.image_composition.blend_map, width, height, render_target, 0);

	D_TOCK(d_ibfv_warp, d_cnt_idx++)



	// finalize
	D_TICK(d_ibfv_warp, d_cnt_idx)

	glBindTexture(GL_TEXTURE_2D, 0);
	GLSLManager::SetRenderTarget(offscreen);
	GLSLManager::SetShader();
	glDisable(GL_TEXTURE_2D);
	glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);

	D_TOCK(d_ibfv_warp, d_cnt_idx++)


	/**
	if(param.image_composition.copy_back == true) {
		for(int i = 0; i < num_enabled_warped_images; i++) {
			for(int j = 0; j < num_elems; j++) {
				tmp_arr2[i][j] = render_target[j];
			}
		}
	}
	/**/
}

#else

/**/
void IBFVRenderer::renderPaintingOnLayer() {
	int num_verts = 3 * param.mesh_info.num_visible_triangles;
	int num_elems = 4 * num_pixels;
	int *coords = param.image_composition.coords;

	unsigned char *render_target;
	unsigned char *tmp_arr1[2], *tmp_arr2[4];

	bool saveflag = false;


	render_target = layers[current_layer]->rendered_image;

	tmp_arr1[0] = layers[current_layer]->initial_stroke_image;
	tmp_arr1[1] = aux[0];


	for(int i = 0; i < 4; i++) {
		tmp_arr2[i] = render_buffer[i];
	}


	// initialize  images
	for(int g = 0; g < 4; g++) {
		for(int i = 0; i < num_elems; i++) {
			tmp_arr2[g][i] = tmp_arr1[0][i];
		}
	}

	glActiveTexture(GL_TEXTURE0);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glEnable(GL_TEXTURE_2D);
	glClear(GL_COLOR_BUFFER_BIT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
 

	// Warp individual 4 images.
	for(int g = 0; g < 4; g++) {
		for(int h = 0; h < param.num_iterations; h++) {
			if(saveflag && g == 2 && h < 2) {
				int idx_3 = 0, idx_4 = 0;
				unsigned char *buffer = new unsigned char[3 * width * height];
				string filename("warp_before_");
				stringstream ss;
				ss.str();
				ss << filename << h << ".ppm";
				for(int i = 0; i < height; i++) for(int j = 0; j < width; j++) {for(int k = 0; k < 3; k++) buffer[idx_3++]=tmp_arr2[g][idx_4++]; idx_4++;}
				My_SSS_SavePPM(width, height, buffer, ss.str().c_str());
				delete[] buffer;
			}

			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, tmp_arr2[g]);

			if(use_gl_nearest_in_warp) {
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			}

			glEnableClientState(GL_VERTEX_ARRAY);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glVertexPointer(2, GL_FLOAT, 0, param.mesh_info.visible_tri_vertex_coords);
			glTexCoordPointer(2, GL_FLOAT, 0, param.mesh_info.visible_tri_vertex_tex_coords[coords[g]]);
			glDrawArrays(GL_TRIANGLES, 0, num_verts);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			glDisableClientState(GL_VERTEX_ARRAY);

			glReadPixels(0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, tmp_arr1[1]);

			if(use_gl_nearest_in_warp) {
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			}

			if(saveflag &&g == 2 && h < 5) {
				int idx_3 = 0, idx_4 = 0;
				unsigned char *buffer = new unsigned char[3 * width * height];
				string filename("warp_after_");
				stringstream ss;
				ss.str();
				ss << filename << h << ".ppm";
				for(int i = 0; i < height; i++) for(int j = 0; j < width; j++) {for(int k = 0; k < 3; k++) buffer[idx_3++]=tmp_arr1[1][idx_4++]; idx_4++;}
				My_SSS_SavePPM(width, height, buffer, ss.str().c_str());
				delete[] buffer;
			}

			overwriteColorBasic(tmp_arr1, bgfg_threshold, width, height, tmp_arr2[g]);
		}
	}

	glDisable(GL_TEXTURE_2D);

	overwriteColor3(tmp_arr2, param.image_composition.blend_map, width, height, render_target, 0);
}
/**/

/**
void IBFVRenderer::renderPaintingOnLayer() {
	int num_verts = 3 * param.mesh_info.num_visible_triangles;
	int num_elems = 4 * num_pixels;
	int *coords = param.image_composition.coords;

	unsigned char *render_target;
	unsigned char *tmp_arr1[2], *tmp_arr2[4];
	unsigned char *read_color_target, *pixel0, *pixel1;
	float hsv0[3], hsv1[3];

	float p0f[3], p1f[3];
	const float frac = 1.f / 255.f;
	float threshold = bgfg_threshold * frac;


	render_target = layers[current_layer]->rendered_image;

	tmp_arr1[0] = layers[current_layer]->initial_stroke_image;
	tmp_arr1[1] = aux[0];


	for(int i = 0; i < 4; i++) {
		tmp_arr2[i] = render_buffer[i];
	}


	// initialize  images
	for(int g = 0; g < 4; g++) {
		for(int i = 0; i < num_elems; i++) {
			tmp_arr2[g][i] = tmp_arr1[0][i];
		}
	}

	glActiveTexture(GL_TEXTURE0);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glEnable(GL_TEXTURE_2D);
	glClear(GL_COLOR_BUFFER_BIT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
 

	int idx = 0;

	// Warp individual 4 images.
	for(int g = 0; g < 4; g++) {
		for(int h = 0; h < param.num_iterations; h++) {
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, tmp_arr2[g]);
			glEnableClientState(GL_VERTEX_ARRAY);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glVertexPointer(2, GL_FLOAT, 0, param.mesh_info.visible_tri_vertex_coords);
			glTexCoordPointer(2, GL_FLOAT, 0, param.mesh_info.visible_tri_vertex_tex_coords[coords[g]]);
			glDrawArrays(GL_TRIANGLES, 0, num_verts);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			glDisableClientState(GL_VERTEX_ARRAY);

			glReadPixels(0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, tmp_arr1[1]);
			//overwriteColorBasic(tmp_arr1, bgfg_threshold, width, height, tmp_arr2[g]);


			if(param.seed_order == 2) {
				idx = 0;

				for(int i = 0; i < num_pixels; i++) {
					if(tmp_arr1[0][idx + 3] * frac < threshold) {
						read_color_target = tmp_arr1[1];
					}
					else if(tmp_arr1[1][idx + 3] * frac < threshold) {
						read_color_target = tmp_arr1[0];
					}
					else {
						/**
						pixel0 = &(tmp_arr1[0][idx]);		// order 1
						pixel1 = &(tmp_arr1[1][idx]);
						RgbHsv(pixel0, hsv0);
						RgbHsv(pixel1, hsv1);
						/**/

						/**
						p0f[0] = pixel0[0] / 255.f;	// this method was a little different... (order 2)
						p0f[1] = pixel0[1] / 255.f;
						p0f[2] = pixel0[2] / 255.f;
						p1f[0] = pixel1[0] / 255.f;
						p1f[1] = pixel1[1] / 255.f;
						p1f[2] = pixel1[2] / 255.f;
						RgbHsv(p0f, hsv0);
						RgbHsv(p1f, hsv1);
						/**/

						/**
						p0f[0] = pixel0[0] * frac;		// order 3
						p0f[1] = pixel0[1] * frac;
						p0f[2] = pixel0[2] * frac;
						p1f[0] = pixel1[0] * frac;
						p1f[1] = pixel1[1] * frac;
						p1f[2] = pixel1[2] * frac;
						RgbHsv(p0f, hsv0);
						RgbHsv(p1f, hsv1);
						/**

						if(StrokeComparator::hsv_comp.eval(hsv0, hsv1) == true) {
							read_color_target = tmp_arr1[0];
						}
						else {
							read_color_target = tmp_arr1[1];
						}
					}

					// sim1
					tmp_arr2[g][idx] = (unsigned char)::clampVal((read_color_target[idx++] / 255.f) * 255.f, 0, 255);
					tmp_arr2[g][idx] = (unsigned char)::clampVal((read_color_target[idx++] / 255.f) * 255.f, 0, 255);
					tmp_arr2[g][idx] = (unsigned char)::clampVal((read_color_target[idx++] / 255.f) * 255.f, 0, 255);
					tmp_arr2[g][idx] = (unsigned char)::clampVal((read_color_target[idx++] / 255.f) * 255.f, 0, 255);

					// sim2
					//tmp_arr2[g][idx] = (unsigned char)::clampVal((read_color_target[idx++] * frac) * 255.f, 0, 255);
					//tmp_arr2[g][idx] = (unsigned char)::clampVal((read_color_target[idx++] * frac) * 255.f, 0, 255);
					//tmp_arr2[g][idx] = (unsigned char)::clampVal((read_color_target[idx++] * frac) * 255.f, 0, 255);
					//tmp_arr2[g][idx] = (unsigned char)::clampVal((read_color_target[idx++] * frac) * 255.f, 0, 255);
				}
			}
		}
	}

	/**
	unsigned char *bg = new unsigned char[3 * width * height];
	int d_idx0 = 0, d_idx1 = 0;
	stringstream tmp_filename;
	for(int i = 0; i < 4; i++) {
		d_idx0 = 0;
		d_idx1 = 0;
		for(int j = 0; j < num_pixels; j++) {
			bg[d_idx0++] = tmp_arr2[i][d_idx1++];
			bg[d_idx0++] = tmp_arr2[i][d_idx1++];
			bg[d_idx0++] = tmp_arr2[i][d_idx1++];
			d_idx1++;
		}

		tmp_filename.str("");

		switch(i) {
			case 0:
				tmp_filename << "warp_(-x).ppm";
				break;

			case 1:
				tmp_filename << "warp_(-y).ppm";
				break;

			case 2:
				tmp_filename << "warp_(+x).ppm";
				break;

			case 3:
				tmp_filename << "warp_(+y).ppm";
				break;
		}

		My_SSS_SavePPM(width, height, bg, tmp_filename.str().c_str());
	}
	delete[] bg;
	/**

	glDisable(GL_TEXTURE_2D);

	overwriteColor3(tmp_arr2, param.image_composition.blend_map, width, height, render_target, 0);
}
/**/

#endif


void IBFVRenderer::overwriteColorBasic(unsigned char **images, unsigned char threshold, int width, int height, unsigned char *output) {
	int idx = 0;
	int num_pixels = width * height;
	unsigned char *pixel0, *pixel1;
	unsigned char *read_color_target;
	float hsv0[3], hsv1[3];

	float p0f[3], p1f[3];
	const float frac = 1.f / 255.f;



	switch(param.seed_order) {
		case 0:
		case 1:
		default:
			for(int i = 0; i < num_pixels; i++) {
				if(images[0][idx + 3] < threshold) {
					read_color_target = images[1];
				}
				else if(images[1][idx + 3] < threshold) {
					read_color_target = images[0];
				}
				else {
					pixel0 = &(images[0][idx]);
					pixel1 = &(images[1][idx]);

					if(StrokeComparator::rgb_comp.eval(pixel0, pixel1) == true) {
						read_color_target = images[0];
					}
					else {
						read_color_target = images[1];
					}	
				}

				output[idx] = read_color_target[idx++];
				output[idx] = read_color_target[idx++];
				output[idx] = read_color_target[idx++];
				output[idx] = read_color_target[idx++];
			}

			break;

		case 2:
			for(int i = 0; i < num_pixels; i++) {
				if(images[0][idx + 3] < threshold) {
					read_color_target = images[1];
				}
				else if(images[1][idx + 3] < threshold) {
					read_color_target = images[0];
				}
				else {
					pixel0 = &(images[0][idx]);
					pixel1 = &(images[1][idx]);

					/**/
					RgbHsv(pixel0, hsv0);
					RgbHsv(pixel1, hsv1);
					/**/

					/**
					p0f[0] = pixel0[0] / 255.f;	// this method was a little different...
					p0f[1] = pixel0[1] / 255.f;
					p0f[2] = pixel0[2] / 255.f;
					p1f[0] = pixel1[0] / 255.f;
					p1f[1] = pixel1[1] / 255.f;
					p1f[2] = pixel1[2] / 255.f;
					RgbHsv(p0f, hsv0);
					RgbHsv(p1f, hsv1);
					/**/

					/**
					p0f[0] = pixel0[0] * frac;
					p0f[1] = pixel0[1] * frac;
					p0f[2] = pixel0[2] * frac;
					p1f[0] = pixel1[0] * frac;
					p1f[1] = pixel1[1] * frac;
					p1f[2] = pixel1[2] * frac;
					RgbHsv(p0f, hsv0);
					RgbHsv(p1f, hsv1);
					/**/

					if(StrokeComparator::hsv_comp.eval(hsv0, hsv1) == true) {
						read_color_target = images[0];
					}
					else {
						read_color_target = images[1];
					}
				}

				output[idx] = read_color_target[idx++];
				output[idx] = read_color_target[idx++];
				output[idx] = read_color_target[idx++];
				output[idx] = read_color_target[idx++];
			}

			break;
	}
}


void IBFVRenderer::overwriteColor3(unsigned char **images, float *weight_map, int width, int height, unsigned char *output, int op) {
	int idx;
	int num_pixels = width * height;
	int num_elems = 4 * num_pixels;
	unsigned char *pixel0, *pixel1;
	unsigned char *img_x[2], *img_y[2];
	unsigned char *read_color_target;
	unsigned char *composite_x, *composite_y;
	float threshold = .1f;		// 0 ~ .5f
	float hsv0[3], hsv1[3];

	threshold = .55f;
	composite_x = aux[0];
	composite_y = aux[1];


	D_SET_COUNTER_VAR(d_cnt_idx, 3 * current_layer)

	switch(op) {
		case 0:						// composite 4 images...
			img_x[0] = images[0];
			img_x[1] = images[2];
			img_y[0] = images[1];
			img_y[1] = images[3];


			D_TICK(d_ibfv_composite, d_cnt_idx)
			overwriteColorBasic(img_x, bgfg_threshold, width, height, composite_x);
			D_TOCK(d_ibfv_composite, d_cnt_idx++)


			D_TICK(d_ibfv_composite, d_cnt_idx)
			overwriteColorBasic(img_y, bgfg_threshold, width, height, composite_y);
			D_TOCK(d_ibfv_composite, d_cnt_idx++)


			idx = 0;

			D_TICK(d_ibfv_composite, d_cnt_idx)
			switch(param.seed_order) {
				case 0:
				case 1:
				default:
					for(int i = 0; i < num_pixels; i++) {
						// take x images
						if(weight_map[i] >= .5f + threshold) {
							read_color_target = composite_x;
						}
						// take y images
						else if(weight_map[i] <= .5f - threshold) {
							read_color_target = composite_y;
						}
						else if(composite_x[idx + 3] < bgfg_threshold) {
							read_color_target = composite_y;
						}
						else if(composite_y[idx + 3] < bgfg_threshold) {
							read_color_target = composite_x;
						}
						else {
							pixel0 = &(composite_x[idx]);
							pixel1 = &(composite_y[idx]);

							if(StrokeComparator::rgb_comp.eval(pixel0, pixel1) == true) {
								read_color_target = composite_x;
							}
							else {
								read_color_target = composite_y;
							}
						}

						output[idx] = read_color_target[idx++];
						output[idx] = read_color_target[idx++];
						output[idx] = read_color_target[idx++];
						output[idx] = read_color_target[idx++];
					}

					break;

				case 2:
					for(int i = 0; i < num_pixels; i++) {
						// take x images
						if(weight_map[i] >= .5f + threshold) {
							read_color_target = composite_x;
						}
						// take y images
						else if(weight_map[i] <= .5f - threshold) {
							read_color_target = composite_y;
						}
						else if(composite_x[idx + 3] < bgfg_threshold) {
							read_color_target = composite_y;
						}
						else if(composite_y[idx + 3] < bgfg_threshold) {
							read_color_target = composite_x;
						}
						else {
							pixel0 = &(composite_x[idx]);
							pixel1 = &(composite_y[idx]);
							RgbHsv(pixel0, hsv0);
							RgbHsv(pixel1, hsv1);

							if(StrokeComparator::hsv_comp.eval(hsv0, hsv1) == true) {
								read_color_target = composite_x;
							}
							else {
								read_color_target = composite_y;
							}	
						}

						output[idx] = read_color_target[idx++];
						output[idx] = read_color_target[idx++];
						output[idx] = read_color_target[idx++];
						output[idx] = read_color_target[idx++];
					}

					break;
			}
			D_TOCK(d_ibfv_composite, d_cnt_idx++)

			break;


		case 1:						// blending x+ and x-, or y+ and y-
			overwriteColorBasic(images, bgfg_threshold, width, height, output);

			break;


		case 2:						// otherwise... (blending x and y images)
									// do nothing...

			break;

		default:
			return;
	}
}


void IBFVRenderer::renderPaintingOnLayerDebug() {
	int num_elems = 4 * num_pixels;
	unsigned char *copy_from, *copy_to;


	copy_from = layers[current_layer]->initial_stroke_image;
	copy_to = layers[current_layer]->rendered_image;


	// initialize
	for(int i = 0; i < num_elems; i++) {
		copy_to[i] = copy_from[i];
	}

	for(int i = 0; i < 4; i++) {
		copy_to = render_buffer[i];

		for(int j = 0; j < num_elems; j++) {
			copy_to[j] = copy_from[j];
		}
	}

	glActiveTexture(GL_TEXTURE0);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glEnable(GL_TEXTURE_2D);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	for(int i = 0; i < param.num_iterations; i++) {
		renderOneIteration();
	}

	glDisable(GL_TEXTURE_2D);
}


void IBFVRenderer::renderOneIteration() {
	int op_type = 0;
	int num_elems = 4 * num_pixels;
	int num_verts = 3 * param.mesh_info.num_visible_triangles;
	int num_enabled_warped_images = param.image_composition.num_enabled_warped_images;
	int *coords = param.image_composition.coords;

	unsigned char *render_target;
	unsigned char *tmp_arr1[2], *tmp_arr2[4];


	if(num_enabled_warped_images < 1 || num_enabled_warped_images == 3 || num_enabled_warped_images > 4) {
		return;
	}


	render_target = layers[current_layer]->rendered_image;
	tmp_arr1[0] = layers[current_layer]->initial_stroke_image;
	tmp_arr1[1] = aux[0];

	for(int i = 0; i < 4; i++) {
		tmp_arr2[i] = render_buffer[i];
	}

	for(int g = 0; g < num_enabled_warped_images; g++) {
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, tmp_arr2[g]);
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glVertexPointer(2, GL_FLOAT, 0, param.mesh_info.visible_tri_vertex_coords);
		glTexCoordPointer(2, GL_FLOAT, 0, param.mesh_info.visible_tri_vertex_tex_coords[coords[g]]);
		glDrawArrays(GL_TRIANGLES, 0, num_verts);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glDisableClientState(GL_VERTEX_ARRAY);

		glReadPixels(0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, tmp_arr1[1]);
		overwriteColorBasic(tmp_arr1, bgfg_threshold, width, height, tmp_arr2[g]);
	}

	if(num_enabled_warped_images == 1) {
		for(int i = 0; i < num_elems; i++) {
			render_target[i] = tmp_arr2[0][i];
		}
	}
	else {
		if(num_enabled_warped_images == 4) op_type = 0;
		else if(num_enabled_warped_images == 2 && coords[0] + 4 == coords[1]) op_type = 1;
		else op_type = 2;

		overwriteColor3(tmp_arr2, param.image_composition.blend_map, width, height, render_target, op_type);
	}

	if(param.image_composition.copy_back == true) {
		for(int i = 0; i < num_enabled_warped_images; i++) {
			for(int j = 0; j < num_elems; j++) {
				tmp_arr2[i][j] = render_target[j];
			}
		}
	}
}


#undef USE_GLSL_IBFV
#undef D_SET_COUNTER_VAR
#undef D_UPDATE_COUNTER_VAR
#undef D_INIT_TIMING
#undef D_TICK
#undef D_TOCK
#undef D_TIME
#undef D_FIN_TIMING
#undef D_USE_DEPTH_IBFV
