#pragma once
#include "stdafx.h"
#include "OpticalFlow.h"

OpticalFlow::OpticalFlow(CvTermCriteria criteria, int window_size, int num_pyramids) {
	this->window_size = window_size;
	this->num_pyramids = num_pyramids;
	this->criteria = criteria;
}


OpticalFlow::~OpticalFlow() {
}


void OpticalFlow::LK(IplImage *img1, IplImage *img2, CvPoint2D32f *result) {
	int idx;
	int width, height;
	int lkwinsize;
	float dx, dy;
	CvMat *velx, *vely;

	assert(img1 != NULL);
	assert(img2 != NULL);
	assert(img1->width == img2->width);
	assert(img1->height == img2->height);
	assert(img1->nChannels == 1);
	assert(img2->nChannels == 1);
	assert(img1->origin == img2->origin);
	assert(result != NULL);


	width = img1->width;
	height = img1->height;
	velx = cvCreateMat(height, width, CV_32FC1);
	vely = cvCreateMat(height, width, CV_32FC1);
	cvSetZero(velx);
	cvSetZero(vely);

	lkwinsize = window_size;

	if(lkwinsize < 1) lkwinsize = 1;
	if(((lkwinsize >> 1) << 1) == lkwinsize) lkwinsize++;
	if(lkwinsize > 15) lkwinsize = 15;

	cvCalcOpticalFlowLK(img1, img2, cvSize(lkwinsize, lkwinsize), velx, vely);


	for(int i = 0; i < height; i++) {
		for(int j = 0; j < width; j++) {
			idx = j + width * i;
			dx = (float)cvGetReal2D(velx, i, j);
			dy = (float)cvGetReal2D(vely, i, j);
			result[idx].x = (float)j + dx;
			result[idx].y = (float)i + dy;
		}
	}

	cvReleaseMat(&velx);
	cvReleaseMat(&vely);
}


void OpticalFlow::HS(IplImage *img1, IplImage *img2, CvPoint2D32f *result) {
	int idx;
	int width, height;
	float dx, dy;
	CvMat *velx, *vely;

	assert(img1 != NULL);
	assert(img2 != NULL);
	assert(img1->width == img2->width);
	assert(img1->height == img2->height);
	assert(img1->nChannels == 1);
	assert(img2->nChannels == 1);
	assert(img1->origin == img2->origin);
	assert(result != NULL);


	width = img1->width;
	height = img1->height;
	velx = cvCreateMat(height, width, CV_32FC1);
	vely = cvCreateMat(height, width, CV_32FC1);
	cvSetZero(velx);
	cvSetZero(vely);

	cvCalcOpticalFlowHS(img1, img2, 0, velx, vely, .5, criteria);


	for(int i = 0; i < height; i++) {
		for(int j = 0; j < width; j++) {
			idx = j + width * i;
			dx = (float)cvGetReal2D(velx, i, j);
			dy = (float)cvGetReal2D(vely, i, j);
			result[idx].x = (float)j + dx;
			result[idx].y = (float)i + dy;
		}
	}

	cvReleaseMat(&velx);
	cvReleaseMat(&vely);
}


void OpticalFlow::PyrLK(IplImage *img1, IplImage *img2, CvPoint2D32f *result) {
	int idx;
	int width, height;
	int elems;
	CvPoint2D32f *frame1_features;
	char *optical_flow_found_feature;

	assert(img1 != NULL);
	assert(img2 != NULL);
	assert(img1->width == img2->width);
	assert(img1->height == img2->height);
	assert(img1->nChannels == 1);
	assert(img2->nChannels == 1);
	assert(img1->origin == img2->origin);
	assert(result != NULL);


	width = img1->width;
	height = img1->height;
	elems = width * height;
	frame1_features = (CvPoint2D32f *)cvAlloc(elems * sizeof(CvPoint2D32f));
	optical_flow_found_feature = (char*)cvAlloc(elems * sizeof(char));

	for(int i = 0; i < height; i++) {
		for(int j = 0; j < width; j++) {
			idx = j + width * i;
			frame1_features[idx].x = (float)j;
			frame1_features[idx].y = (float)i;
		}
	}

	cvCalcOpticalFlowPyrLK(
		img1, 
		img2, 
		NULL, 
		NULL,
		frame1_features, 
		result, 
		elems, 
		cvSize(window_size, window_size), 
		num_pyramids, 
		optical_flow_found_feature, 
		NULL, 
		cvTermCriteria(CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 1, 1), 
		0
	);

	cvFree((void**)&frame1_features);
	cvFree((void**)&optical_flow_found_feature);
}



void OpticalFlow::LK(IplImage *img1, IplImage *img2, float *result) {
	int idx = 0;
	int num_elems = img1->width * img1->height;
	CvPoint2D32f *tmp_result = (CvPoint2D32f *)cvAlloc(num_elems * sizeof(CvPoint2D32f));
	
	LK(img1, img2, tmp_result);

	for(int i = 0; i < num_elems; i++) {
		result[idx++] = tmp_result[i].x;
		result[idx++] = tmp_result[i].y;
	}

	cvFree((void**)&tmp_result);
}


void OpticalFlow::HS(IplImage *img1, IplImage *img2, float *result) {
	int idx = 0;
	int num_elems = img1->width * img1->height;
	CvPoint2D32f *tmp_result = (CvPoint2D32f *)cvAlloc(num_elems * sizeof(CvPoint2D32f));
	
	HS(img1, img2, tmp_result);

	for(int i = 0; i < num_elems; i++) {
		result[idx++] = tmp_result[i].x;
		result[idx++] = tmp_result[i].y;
	}

	cvFree((void**)&tmp_result);
}


void OpticalFlow::PyrLK(IplImage *img1, IplImage *img2, float *result) {
	int idx = 0;
	int num_elems = img1->width * img1->height;
	CvPoint2D32f *tmp_result = (CvPoint2D32f *)cvAlloc(num_elems * sizeof(CvPoint2D32f));
	
	PyrLK(img1, img2, tmp_result);

	for(int i = 0; i < num_elems; i++) {
		result[idx++] = tmp_result[i].x;
		result[idx++] = tmp_result[i].y;
	}

	cvFree((void**)&tmp_result);
}

