#pragma once
#include "StdAfx.h"
#include "PaintingStyleField.h"
#include <list>
#include <fstream>

using namespace std;


const float PaintingStyleField::mp_default_stroke_spacing = 6.f;
//const float PaintingStyleField::mp_default_stroke_spacing = 5.f;


PaintingStyleField::PaintingStyleField(int width, int height)   {
	PaintingRegion *pr;
	PaintingPreset p;
	int idx;
	int bg_id = 0;
	int num_elems = width * height;

	this->width = width;
	this->height = height;
	style = new PaintingStyleNode[num_elems];
    style_floats = new PPInterpolatable[num_elems];



	// first put only one region... (background)
	global_region_id = bg_id;

	pr = new PaintingRegion(global_region_id++);
	p = pr->getKey(0);

	for(int i = 0; i < height; i++) {
		for(int j = 0; j < width; j++) {
			idx = j + width * i;

			style[idx].region_id = bg_id;
			style[idx].preset = p;
			pr->points.push_back(pair<int, int>(j, i));
		}
	}

	regions.insert(pair<int, PaintingRegion*>(pr->getID(), pr));


	// allocate buffer for design element propagation
	num_frames = 1;
	setDesignElemHolderSizeForRegion(num_frames);

	mp_stroke_spacing = mp_default_stroke_spacing;


	diffuse = true;						// enable diffusion  (I've set up this variable for debug purpose...)

	// diffusion setting...
	psf_blur.blurWithPyramid(true);										// enable pyramid method
	psf_blur.enableConvergenceCheck(true);								// check convergence at each pyramid level
	psf_blur.setBlurMethod(Diffusion<PPInterpolatable>::GAUSS_SEIDEL);	// use Gauss Seidel method
	//psf_blur.setConvergenceCriteria(0.01);								// use this function for convergence error control
	diffuse_num_itr = 1;												// Pyramid level keep iterates at each level until values converge... so set 1 for this.
	diffuse_stepsize = 10.f;											// use a large timestep for Gauss Seidel
}


PaintingStyleField::~PaintingStyleField() {
	map<int, PaintingRegion*>::iterator it;
	delete[] style;
    delete[] style_floats;

	for(it = regions.begin(); it != regions.end(); it++) {
		delete (*it).second;
	}
}


float PaintingStyleField::getStrokeSpacingForMultiStyle() {
	return mp_stroke_spacing;
}


void PaintingStyleField::setStrokeSpacingForMultiStyle(float val) {
	mp_stroke_spacing = ((val >= 1) ? val : 1);
}


void PaintingStyleField::setMask(bool *mask) {
	int num_elems = width * height;

	if(mask == NULL) return;

	for(int i = 0; i < num_elems; i++) {
		style[i].mask = mask[i];
	}
}


void PaintingStyleField::setUniformMask(bool mask) {
	int num_elems = width * height;

	for(int i = 0; i < num_elems; i++) {
		style[i].mask = mask;
	}
}


void PaintingStyleField::setStyle(PaintingPreset preset) {
	int num_elems = width * height;

	for(int i = 0; i < num_elems; i++) {
		if(style[i].mask == true) style[i].preset = preset;
	}
}


void PaintingStyleField::blurStyleParameters(int num_itr, float dt) {
	int num_elems = width * height;
    bool no_style_defined = true;
	PaintingPreset *pp1;
	list< pair<int, int> > *points;
	list< pair<int, int> >::iterator itr_list;
	map<int, PaintingRegion*>::iterator itr_map;


	if(width <= 0 || height <= 0) return;

	for(itr_map = regions.begin(); itr_map != regions.end(); itr_map++) {
		if((*itr_map).second->getNumKeys() > 0) {
			no_style_defined = false;
			break;
		}
	}

	// no interpolation is necessary if no region has its own style.
	if(no_style_defined == true) return;

	// apply mask
    // apply mask (unmask is not necessary...)
	for(itr_map = regions.begin(); itr_map != regions.end(); itr_map++) {
		points = &((*itr_map).second->points);

		if((*itr_map).second->getNumKeys() > 0) {
			for(itr_list = points->begin(); itr_list != points->end(); itr_list++) {
				style_floats[itr_list->first + width * itr_list->second].mask = false;
			}
		}
		else {
			// set values for uninterpolated.
			for(itr_list = points->begin(); itr_list != points->end(); itr_list++) {
                   style_floats[itr_list->first + width * itr_list->second].mask = true;

				// set values for uninterpolated.
			
				pp1 = &(style[itr_list->first + width * itr_list->second].preset);
				pp1->stroke_gridsize = mp_stroke_spacing;
				pp1->color_composition_method = (int)mp_color_composition;
				pp1->alpha = mp_alpha;
				pp1->post_process_operation = mp_postprocessing_type;
			}
		}
	}


	for(int i = 0; i < num_elems; i++) {
		style_floats[i].pack(style[i].preset);
	}

	// blur (heat diffusion)
	psf_blur.blur(style_floats, width, height, dt, num_itr);

	for(int i = 0; i < num_elems; i++) {
		style_floats[i].unpack(style[i].preset);
	}
}


PaintingRegion* PaintingStyleField::getRegion(int x, int y) {
	map<int, PaintingRegion*>::iterator it;

	it = regions.find(style[x + width * y].region_id);

	if(it == regions.end()) return NULL;
	else return it->second;
}


PaintingRegion* PaintingStyleField::createNewRegion(vector< pair<int, int> > &path, int time) {
	int idx;
	int x, y;
	int x0, y0;
	int region_id;
	bool flag;
	PaintingRegion *opr;
	PaintingRegion *npr;
	map<int, PaintingRegion*>::iterator itm;
	list< pair<int, int> >::iterator itl;
	pair<int, int> seed;


	// no path is passed... return
	if(path.size() <= 0) return NULL;

	x0 = path.at(0).first;
	y0 = path.at(0).second;

	// look for the region ID to split
	if(0 <= x0 && x0 < width && 0 <= y0 && y0 < height) {
		region_id = style[x0 + width * y0].region_id;
	}
	else {
		return NULL;
	}

	// grab a region object
	itm = regions.find(region_id);

	if(itm == regions.end()) {
		return NULL;
	}

	// the current region
	opr = itm->second;

	// new region
	npr = new PaintingRegion(global_region_id++);

	// draw a closed curve on style field
	drawPath(path, npr->getID());

	flag = NULL;

	// look for a starting point for floodfill
	for(int i = 0; i < 3 && flag == false; i++) {
		y = y0 - 1 + i;

		if(0 <= y && y < height) {
			for(int j = 0; j < 3 && flag == false; j++) {
				x = x0 - 1 + j;

				if(0 <= x && x < width) {
					if(style[x + width * y].region_id == region_id) {
						seed.first = x;
						seed.second = y;
						flag = true;
					}
				}
			}
		}
	}


	// hope this never happens...
	if(flag == false) {
		delete npr;
		return NULL;
	}


	// flood fill
	floodFill(seed, npr->getID());


	// update a point set of each region
	for(itl = opr->points.begin(); itl != opr->points.end();) {
		idx = itl->first + width * itl->second;

		if(style[idx].region_id == npr->getID()) {
			npr->points.push_back(*itl);
			itl = opr->points.erase(itl);
		}
		else {
			itl++;
		}
	}

	// copy painting style from the current region
	if(opr->getNumKeys() > 0) {
		npr->copyKeys(*opr);
	}

	// allocate buffer for design element propagation.
	npr->setDesignElemHolderSize(num_frames);


	regions[npr->getID()] = npr;


	return npr;
}


bool PaintingStyleField::applyPaintingForRegion(PaintingRegion* pr, PaintingPreset p) {
	int idx;
	map<int, PaintingRegion*>::iterator itm;
	list< pair<int, int> >::iterator itl;


	if(pr == NULL) return false;

	itm = regions.find(pr->getID());

	if(itm == regions.end()) {
		return false;
	}
	else {
		for(itl = pr->points.begin(); itl != pr->points.end(); itl++) {
			idx = itl->first + width * itl->second;
			style[idx].preset = p;
		}

		if(diffuse == true) {
			blurStyleParameters(diffuse_num_itr, diffuse_stepsize);
		}
	/////////////////////////////////////////////////////////////////////////////////////////////////
	//output
    /////////////////////////////////////////////////////////////////////////////////////////////////

	/*output++;
	CString num;
	num.Format(_T("frame_para%d.txt"),output);
	ofstream outfile(num);
	int idx2;
	

	for(int i=0;i<height;i++) 
	{
		for(int j=0;j<width;j++)
		{
			
			idx2 = j + width * i;

			int regionID, layer ;
			
			regionID=this->style[idx2].region_id;
			layer=this->style[idx2].preset.num_layers;

			outfile<<"( "<<j<<","<<i<<","<<regionID<<","<<layer<<"), ";
            
		}
		outfile<<";"<<endl;
}
    


	outfile.close();*/
	///////////////////////////////////////////////////////////////////////////////////////

		return true;
	}
}


bool PaintingStyleField::updatePaintingOnField(int time) {
	int idx;
	PaintingPreset p;
	PaintingRegion *pr;
	map<int, PaintingRegion*>::iterator itm;
	list< pair<int, int> >::iterator itl;


	for(itm = regions.begin(); itm != regions.end(); itm++) {
		// assign style parameters to each pixel of each region
		if((*itm).second->getNumKeys() > 0) {
			pr = (*itm).second;
			p = pr->getKey(time);

			for(itl = pr->points.begin(); itl != pr->points.end(); itl++) {
				idx = itl->first + width * itl->second;
				style[idx].preset = p;
			}
		}
	}

	if(diffuse == true) {
		blurStyleParameters(diffuse_num_itr, diffuse_stepsize);
	}

	return true;
}


size_t PaintingStyleField::getNumRegions() {
	return regions.size();
}


void PaintingStyleField::drawPath(vector< pair<int, int> > &path, int region_id) {
	int pt0[2], pt1[2];
	int delta[2];
	int sign[2];
	int x0, y0;
	int p;
	int num_itr;


	for(size_t i = 0; i < path.size(); i++) {
		pt0[0] = path.at(i).first;
		pt0[1] = path.at(i).second;
		pt1[0] = path.at((i < path.size() - 1) ? (i + 1) : 0).first;
		pt1[1] = path.at((i < path.size() - 1) ? (i + 1) : 0).second;


		if(pt0[0] == pt1[0] && pt1[0] == pt1[1]) continue;

		x0 = pt0[0];
		y0 = pt0[1];

		if(0 <= x0 && x0 < width && 0 <= y0 && y0 < height) {
			style[(int)x0 + width * (int)y0].region_id = region_id;
		}


		// Bresenham's Algorithm...

		for(int j = 0; j < 2; j++) {
			delta[j] = pt1[j] - pt0[j];
			sign[j] = ((delta[j] < 0) ? -1 : 1);
			delta[j] *= sign[j];
		}

		// dx > dy
		if(delta[0] >= delta[1]) {
			num_itr = delta[0];
			p = 2 * delta[1] - delta[0];

			for(int j = 1; j <= num_itr; j++) {
				if(p < 0) {
					x0 += sign[0];
					p += 2 * delta[1];
				}
				else {
					x0 += sign[0];
					y0 += sign[1];
					p += 2 * (delta[1] - delta[0]);
				}

				if(0 <= x0 && x0 < width && 0 <= y0 && y0 < height) {
					style[x0 + width * y0].region_id = region_id;
				}
			}
		}
		else {
			num_itr = delta[1];
			p = 2 * delta[0] - delta[1];

			for(int j = 1; j <= num_itr; j++) {
				if(p < 0) {
					y0 += sign[1];
					p += 2 * delta[0];
				}
				else {
					x0 += sign[0];
					y0 += sign[1];
					p += 2 * (delta[0] - delta[1]);			
				}

				if(0 <= x0 && x0 < width && 0 <= y0 && y0 < height) {
					style[x0 + width * y0].region_id = region_id;
				}
			}
		}
	}
}


void PaintingStyleField::floodFill(pair<int, int> seed, int region_id) {
	int w, e, t;
	int target_id;
	int idx;
	list< pair<int, int> > queue;
	pair<int, int> node;

	if(seed.first < 0 || seed.first >= width || seed.second < 0 || seed.second >= height) return;

	// flood fill
	queue.push_back(seed);

	target_id = style[seed.first + width * seed.second].region_id;


	while(queue.empty() == false) {
		node = queue.front();
		queue.pop_front();
		idx = node.first + width * node.second;

		if(style[idx].region_id == region_id) {
			continue;
		}
		else {
			w = 0;
			e = 0;

			while(node.first - w - 1 >= 0 && style[idx - w - 1].region_id != region_id) {
				w++;
			}

			while(node.first + e + 1 < width && style[idx + e + 1].region_id != region_id) {
				e++;
			}

			t = w + e + 1;

			for(int i = 0; i < t; i++) {
				if(style[idx - w + i].region_id == target_id) {
					style[idx - w + i].region_id = region_id;

					if(node.second > 0) {
						queue.push_back(pair<int, int>(node.first - w + i, node.second - 1));
					}

					if(node.second < height - 1) {
						queue.push_back(pair<int, int>(node.first - w + i, node.second + 1));
					}
				}
			}
		}
	}
}


PaintingRegion* PaintingStyleField::selectRegion(int x, int y) {
	map<int, PaintingRegion*>::iterator itm;

	if(x < 0 || x >= width || y < 0 || y >= height) return NULL;

	itm = regions.find(style[x + width * y].region_id);

	if(itm != regions.end()) {
		return itm->second;
	}
	else {
		return NULL;
	}
}


PaintingRegion* PaintingStyleField::selectRegion(int region_id) {
	map<int, PaintingRegion*>::iterator itm;

	itm = regions.find(region_id);

	if(itm != regions.end()) {
		return itm->second;
	}
	else {
		return NULL;
	}
}


void PaintingStyleField::flatten(bool retain_style) {
	int idx;
	int bg_id = 0;
	map<int, PaintingRegion*>::iterator itm;
	PaintingRegion *pr;
	PaintingPreset retained,p;


	if(retain_style == true) {
		for(itm = regions.begin(); itm != regions.end(); itm++) {
			if(itm->second->getNumKeys() > 0) {
				retained = itm->second->getKey(0);
				break;
			}
		}

		// set false if no region has its own style
		if(itm == regions.end()) retain_style = false;
	}
	itm = regions.begin();

	// erase all painting regions
	while(itm != regions.end()) {
		delete itm->second;
		itm = regions.erase(itm);
	}

	regions.clear();

	global_region_id = bg_id;

	// reset region id counter, and create a single region
	pr = new PaintingRegion(global_region_id++);

    if(retain_style == true) {
		pr->setKey(retained, 0);
	}

	p = pr->getKey(0);

	for(int i = 0; i < height; i++) {
		for(int j = 0; j < width; j++) {
			idx = j + width * i;
			style[idx].region_id = bg_id;
			style[idx].preset = p;
			pr->points.push_back(pair<int, int>(j, i));
		}
	}

	regions.insert(pair<int, PaintingRegion*>(pr->getID(), pr));

	setDesignElemHolderSizeForRegion(num_frames);
}


int PaintingStyleField::getMaxLayers() {
	int num_layers = 0, tmp_num_layers;
	map<int, PaintingRegion*>::iterator itm;


	for(itm = regions.begin(); itm != regions.end(); itm++) {

		// just take a maximum of values at regions where styles are specified
		if(itm->second->getNumKeys() > 0) {
			tmp_num_layers = itm->second->getCurrentKey().num_layers;

			if(num_layers < tmp_num_layers) num_layers = tmp_num_layers;
		}
	}

	return num_layers;
}


void PaintingStyleField::setRegion(unsigned char *binary_map, int width, int height) {
	int wl, hl;
	int idx, idx2;
	size_t num_regions;
	map<int, PaintingRegion*>::iterator itm;
	PaintingRegion *pr[2];

	num_regions = regions.size();

	if(num_regions >= 2) {
		itm = regions.begin();
		pr[0] = (*itm).second;
		itm++;
		pr[1] = (*itm).second;
		itm++;

		for( ; itm != regions.end(); ) {
			itm = regions.erase(itm);
		}
	}
	// assume the case is 1
	else {
		pr[0] = regions.begin()->second;
		pr[1] = new PaintingRegion(global_region_id++);

		regions[pr[1]->getID()] = pr[1];
	}

	pr[0]->points.clear();
	pr[1]->points.clear();


	wl = (this->width < width) ? this->width : width;
	hl = (this->height < height) ? this->height : height;
	hl = this->height - hl;


	// assign to the region 0 for undefined pixels...
	for(int i = 0; i < hl; i++) {
		for(int j = 0; j < this->width; j++) {
			idx2 = j + this->width * i;

			pr[0]->points.push_back(pair<int, int>(j, i));
			style[idx2].region_id = pr[0]->getID();
		}
	}


	for(int i = hl; i < this->height; i++) {
		// assign to regions according to a binary map.
		for(int j = 0; j < wl; j++) {
			idx = j + width * (height - 1 - (i - hl));
			idx2 = j + this->width * i;

			if(binary_map[idx] == 0) {
				pr[0]->points.push_back(pair<int, int>(j, i));
				style[idx2].region_id = pr[0]->getID();
			}
			else {
				pr[1]->points.push_back(pair<int, int>(j, i));
				style[idx2].region_id = pr[1]->getID();
			}
		}

		// assign to the region 0 for undefined pixels...
		for(int j = wl; j < this->width; j++) {
			idx2 = j + this->width * i;

			pr[0]->points.push_back(pair<int, int>(j, i));
			style[idx2].region_id = pr[0]->getID();
		}
	}
}


float PaintingStyleField::findMinStrokeSize(int time) {
	pair<int, int> pt;
	float size = -1.f, tmp_size;
	PaintingPreset p;
	map<int, PaintingRegion*>::iterator itm;


	/**
	FILE *fp = fopen("test.txt", "a"); int c = 0;
	fprintf(fp, "Frame %d\n", time);
	for(itm = regions.begin(); itm != regions.end(); itm++) {
		if(itm->second->points.size() > 0) {
			pt = itm->second->points.front();
			p = style[pt.first + width * pt.second].preset;

			fprintf(fp, "Region %2d: %d, %d\n", c++, itm->second->points.size(), p.stroke_size);
		}
	}
	fclose(fp);
	/**/

	// take a look at the first element rather than computing avarage... (for efficiency)
	for(itm = regions.begin(); itm != regions.end(); itm++) {
		if(itm->second->points.size() > 0) {
			pt = itm->second->points.front();
			p = style[pt.first + width * pt.second].preset;
			//p = itm->second->getKey(time);		// <-- this won't work for style diffusion...
			tmp_size = p.stroke_size;

			if(size < 0 || tmp_size < size) size = tmp_size;
		}
	}

	if(size < 0) size = 0.f;

	return size;
}


map<int, PaintingRegion*>::iterator PaintingStyleField::getBegin() {
	return regions.begin();
}


map<int, PaintingRegion*>::iterator PaintingStyleField::getEnd() {
	return regions.end();
}


PaintingPreset PaintingStyleField::getStyleAt(int x, int y) {
	if(x < 0 || x >= width || y < 0 || y >= height) {
		return PRESET_NONE;
	}
	else {
		return style[x + width * y].preset;
	}
}

void PaintingStyleField::SetSegmentationRegion(unsigned int* RegionIDArray)
{
	int idx;
	int idx2;
	map<int, PaintingRegion*>::iterator itm;
	size_t num_regions;
    unsigned int currentID;
	
	
	global_region_id=0;

	//clear the map first
	num_regions = regions.size();
	if(num_regions)
	{
		for(itm=regions.begin(); itm != regions.end();itm++)
		{
			PaintingRegion *pr;
			pr= (*itm).second;
			delete pr;

		}
		regions.clear();
	}
	//put the segmentation region into the map
 
	for(int i=0;i<height;i++)
	{
		for(int j=0;j<width;j++)
		{
			idx = j + width * (height  - i-1);
			idx2 = j + width * i;
			currentID=RegionIDArray[idx2];
			if(currentID==regions.size())
			{
				//create a new region
				PaintingRegion *temp=new PaintingRegion(global_region_id++);
				regions.insert(pair<int,PaintingRegion*>(temp->getID(),temp));
				//delete [] temp;
			}
			
			regions[currentID]->points.push_back(pair<int, int>(j, height  - i-1));				
			
				style[idx].region_id=regions[currentID]->getID();		

		}
	}
}


bool PaintingStyleField::setLabelsToRegions(const RegionTransformation &new_regions) {
	int idx, idx2;
	bool ordered;
	const int max_regions = 12;
	size_t num_regions;
	map<int, PaintingRegion*>::iterator itm;

	PaintingRegion *pr[max_regions];				//set the number of regions plus one for boundary
    unsigned int currentID;							//set the currentID of the region	


	if(new_regions.imgdat == NULL) return false;

	if(new_regions.imgdat->height != height || new_regions.imgdat->width != width) {
		// dimension does not match...
		return false;
	}

	num_regions = regions.size();

	//create new regions

	idx = 0;
	ordered = true;
	global_region_id = num_regions;

	for(itm = regions.begin(); itm != regions.end(); itm++) {
		if(idx < max_regions) {
			pr[idx] = itm->second; 

			if(idx != pr[idx]->getID()) ordered = false;

			pr[idx]->setID(idx);
			idx++;
		}
		else {
			delete itm->second;
			ordered = false;
		}
	}

	while(idx < max_regions) {
		ordered = false;
		pr[idx] = new PaintingRegion(idx);
		idx++;
	}

	regions.clear();

	for(int i = 0; i < max_regions; i++) {
		regions[i] = pr[i]; 
	}


	// regenerate design elements
	//if(ordered == true) {
	if(ordered == false) {
		setDesignElemHolderSizeForRegion(num_frames);
	}

	// clear the points
	for(int j = 0; j < max_regions; j++) {
		pr[j]->points.clear();
	}
   
	//assign the regions according label 	
	for(int i = 0; i < height;i++) {
		for(int j = 0; j < width; j++) {
			idx = j + width * (height - i - 1);
			idx2 = j + width * i;
			currentID = new_regions.imgdat->labels[idx];

			pr[currentID]->points.push_back(pair<int,int>(j, i));
			style[idx2].region_id = pr[currentID]->getID();    
		}
	}


	/**
	fstream err_file("regions.txt", ios_base::out);
	for(itm = regions.begin(); itm != regions.end(); itm++) {
		err_file << itm->second->getID() << " " << itm->second->points.size() << endl;
	}
	err_file.close();
	/**/

	return true;
}




void PaintingStyleField::getRegionArray(vector<PaintingRegion*> &region_array) {
	size_t vec_size;
	map<int, PaintingRegion*>::reverse_iterator ritr = regions.rbegin();

	// resize based on maximum region ID
	vec_size = ritr->second->getID() + 1;

	region_array.resize(vec_size);

	for(size_t i = 0; i < vec_size; i++) {
		region_array[i] = NULL;
	}

	for(; ritr != regions.rend(); ritr++) {
		region_array[ritr->first] = ritr->second;
	}
}


// Returns a pointer to a painting region based on Region ID passed as argument.
// It gives you NULL if there is no PaintingRegion associated to the 'id'.
PaintingRegion* PaintingStyleField::getRegion(int id) {
	map<int, PaintingRegion*>::iterator itr = regions.find(id);

	return (itr == regions.end()) ? NULL : itr->second;
}



bool PaintingStyleField::propagateDesignElements(SingularElementList *se, RegularElementList *re) {
	return propagateDesignElements(se->singularelements, se->num_singularelements, re->regularelements, re->num_regularelements);
}


bool PaintingStyleField::propagateDesignElements(SingularElement **se, size_t num_se, RegularElement **re, size_t num_re) {
	bool ret = true;
	map<int, PaintingRegion*>::iterator itr;

	for(itr = regions.begin(); itr != regions.end(); itr++) {
		ret &= itr->second->propagateDesignElems(se, num_se, re, num_re);
	}

	return ret;
}



void PaintingStyleField::setNumFrames(int num_frames, const string &file_path) {
	if(num_frames <= 0) num_frames = 1;

	this->num_frames = num_frames;
	setDesignElemHolderSizeForRegion(num_frames, file_path);
}


void PaintingStyleField::setDesignElemHolderSizeForRegion(int size, const string &file_path) {
	map<int, PaintingRegion*>::iterator itr;

	if(num_frames <= 0) num_frames = 1;

	for(itr = regions.begin(); itr != regions.end(); itr++) {
		itr->second->setDesignElemHolderSize(num_frames, file_path);
	}
}


void PaintingStyleField::getDEInfo(vector<const DEInfo*> &design_elements, int time) {
	size_t vec_size;
	map<int, PaintingRegion*>::reverse_iterator ritr = regions.rbegin();

	// resize based on maximum region ID
	vec_size = ritr->second->getID() + 1;

	design_elements.resize(vec_size);

	for(size_t i = 0; i < vec_size; i++) {
		design_elements[i] = NULL;
	}

	for(; ritr != regions.rend(); ritr++) {
		design_elements[ritr->first] = ritr->second->getDesignElems(time);
	}
}
