#pragma once
#include "PaintingPreset.h"
#include "PaintingStyleNode.h"
#include "PaintingRegion.h"
#include "RegionTransformation.h"
#include "Diffusion.h"
#include "tendes_plane.h"
#include <map>
#include <vector>
//#include <fstream>
//#include <iostream>
//#include <sstream>

using namespace std;


/*
	This class wraps multi-style painting operations. multiple regions can be created, different painting styles are assigned to each of them, or can be reduced to one single region.

	PaintingStyleNode is a primitive structure defined on each pixel of PaintingStyleField.
	PaintingRegion holds a collection of PaintingStyleNode objects.
	Origin is assumed to be at lower left corner.


	Note: we are now making multi-style operations in region-based approach, which makes me motivated to rewrite/merge some classes to integrate with 
*/

class PaintingStyleField {
private:
	int global_region_id;
	map<int, PaintingRegion*> regions;

	int width, height;
	int num_frames;
	PaintingStyleNode *style;
	PPInterpolatable *style_floats;
	Diffusion<PPInterpolatable> psf_blur;
	bool diffuse;
	int diffuse_num_itr;
	float diffuse_stepsize;


	void drawPath(vector< pair<int, int> > &path, int region_id);
	void floodFill(pair<int, int> seed, int region_id);


	void setDesignElemHolderSizeForRegion(int size, const string &file_path = string());			// allocate buffer for design element propagation for each region


	// multistyle painting constants...

private:
	float mp_stroke_spacing;								// use getStrokeSpacingForMultiStyle() to get the minimum stroke spacing among regions with different painting styles...

public:
	static const int mp_color_composition = 1;		// 0: alpha_blend, 1:overwrite (it doesn't work if set to 0...)
	static const int mp_alpha = 255;
	static const int mp_stroke_color = 0;
	static const int mp_postprocessing_type = 1;

	static const float mp_default_stroke_spacing;	// 6.0f;

	float getStrokeSpacingForMultiStyle();
	void setStrokeSpacingForMultiStyle(float val);




public:
	PaintingStyleField(int width, int height);
	~PaintingStyleField();

	int getWidth() { return width; }
	int getHeight() { return height; }
	PaintingStyleNode* getStyleNode() { return style; }						// direct access to style nodes

	void setMask(bool *mask);
	void setUniformMask(bool mask);
	void setStyle(PaintingPreset preset);

	void setNumFrames(int num_frames, const string &file_path = string());
	int getNumFrames() { return num_frames; }

	// multi-painting for styles
	void blurStyleParameters(int num_itr, float dt);

	// multi-painting for design elements.
	bool propagateDesignElements(SingularElementList *se, RegularElementList *re);
	bool propagateDesignElements(SingularElement **se, size_t num_se, RegularElement **re, size_t num_re);
	void getDEInfo(vector<const DEInfo*> &design_elements, int time);

	map<int, PaintingRegion*>::iterator getBegin();
	map<int, PaintingRegion*>::iterator getEnd();
	PaintingRegion* getRegion(int id);
	PaintingRegion* getRegion(int x, int y);
	void getRegionArray(vector<PaintingRegion*> &region_array);

	PaintingRegion* createNewRegion(vector< pair<int, int> > &path, int time);
	bool applyPaintingForRegion(PaintingRegion* pr, PaintingPreset p);
	bool updatePaintingOnField(int time);
	size_t getNumRegions();


	PaintingPreset getStyleAt(int x, int y);

	PaintingRegion* selectRegion(int x, int y);
	PaintingRegion* selectRegion(int region_id);
	void flatten(bool retain_style = false);							// collapse into one single region

	int getMaxLayers();													// return the maximum layers to be used for multi-style painting
	float findMinStrokeSize(int time);									// return the minimum stroke size for multi-style painting for video

	void setRegion(unsigned char *binary_map, int width, int height);	// create two regions with different ID (...note that, two disconnected regions but sharing the same ID are still considered as one single region)
	void SetSegmentationRegion(unsigned int* RegionIDArray);
	bool setLabelsToRegions(const RegionTransformation &new_regions);
};
