#pragma once

#include "cv.h"
#include "highgui.h"
#define _USE_MATH_DEFINES
#include <math.h>
#include <stdio.h>
#include <stdlib.h> 
#include <vector>
using namespace std;
//#define M_PI 3.14159 
class RegionTransformation
{
public:
	RegionTransformation(void);
public:
	~RegionTransformation(void);

	//importing data structure

	/*
 ***********************************************************************************
 *	region ellipse structure
 *	------------------------
 *	-> was created given the first 2 orders of the moment of the region
 *
 *	(x,y):				center of the ellipse, also called center of mass
 *	theta:				orientation of the ellipse in degree
 *	minor|majorAxis:	axis length of the ellipse
 *	nPixels:			number of pixels in the region
 ***********************************************************************************
 */
typedef
struct ellipse
{
	union
	{
		struct{double x,y;};					/* center of the ellipse, also called center of mass */
		double xy[2];
	};

	double theta;								/* orientation in degree */

	union
	{
		struct{double majorAxis,minorAxis;};	/* axis lengths */
		double axis[2];
	};
	int nPixels;								/* number of pixels in the region */
}
ellipse;

/*
 ***********************************************************************************
 *	ellipse tranformation structure
 *	-------------------------------
 *	-> transformation parameters from an ellipse of a region i in frame K 
 *	to the ellipse of the same region i in frame K+1
 *
 *	tx,ty:				translation between ellipses
 *	dTheta:				rotation between ellipses
 *	sM,sm:				scale along the major and minor axis
 *	s:					ratio between the number of pixels
 *	h:					resulting 3x3 transformation matrix from 
 *						the previous parameters
 ***********************************************************************************
 */
typedef 
struct ellipseTransformation
{
	double tx,ty;		/* translation */
	double sM,sm,s;		/* scale */
	double dTheta;		/* rotation */
	double h[9];		/* 3x3 transformation matrix */
}
ellipseTransf;

/*
 ***********************************************************************************
 *	region data structure
 *	---------------------
 *	-> contain both ellipses at frame K and K+1, the transformation parameters
 *	between ellipses, the mapping between pixels in a region i at frame K and K+1
 *
 *	id:					region id
 *	eCur,eNext:			ellipses at frame K and K+1
 *	eTransf:			the transformation parameters between ellipses
 *	h:					mapping between pixels in a region i at frame K and K+1
 ***********************************************************************************
 */
typedef
struct regionDat
{
	int id;						/* region id */
	ellipse eCur,eNext;			/* ellipses at frame K and K+1 */
	ellipseTransf eTransf;		/* ellipse transformation */
	double h[9];				/* pixel mapping */
}
regionDat;

/*
 ***********************************************************************************
 *	image data structure
 *	--------------------
 *	-> contain the image dimensions, the region data and the pixel labels
 *
 *	width,height:		width and height of the image
 *	regions:			array of region data
 *	nRegions:			number of regions
 *	labels:				array of pixel labels
 ***********************************************************************************
 */
typedef
struct imageDat
{
	int width,height;		/* image dimension */
	regionDat* regions;		/* array of regions */
	int nRegions;			/* #regions */
	int* labels;			/* array of labels*/
}
imageDat;




CvPoint2D32f getTranfPixel(CvPoint2D32f p, double* h);
CvScalar getIdCol(int id);
void displayVec(IplImage* img, CvPoint p, CvPoint q, CvScalar line_color, int line_thickness, int dispArrow);
bool loadDat(const char* filename);
int releaseImageDat(imageDat** imDat);
void getDat(const char* filename);
imageDat* imgdat;
unsigned char*DatImage;
bool DatImageDefined;
vector <imageDat> data;
int getID(int x, int y);

private:
	IplImage *_img;
};
