#pragma once
#include "StdAfx.h"
#include "SplineStroke.h"
#include <math.h>


struct SplineCoefficients {
	float scale_factor;
	int m0[16];
	int mi[16];
	int mn[16];
};


const SplineCoefficients *sc = NULL;


// Catmull-Rom Splne
//http://tpot.jpn.ph/t-pot/program/2_3rdcurve/index.html
const SplineCoefficients SC_CATMULL_ROM = {
	.5f, 

	{
	 0,  0,  0,  0, 
	 0,  1, -2,  1, 
	 0, -3,  4, -1, 
	 0,  2,  0,  0, 
	}, 

	{
	-1,  3, -3,  1, 
	 2, -5,  4, -1, 
	-1,  0,  1,  0, 
	 0,  2,  0,  0, 
	}, 

	 {
	 0,  0,  0,  0, 
	 1, -2,  1,  0, 
	-1,  0,  1,  0, 
	 0,  2,  0,  0, 
	 }
};


const SplineCoefficients SC_UNIFORM_B_SPLINE = {
	1.f / 6.f, 

	{
	 0,  2, -3,  1, 
	 0, -3,  3,  0, 
	 0, -3,  3,  0, 
	 0,  5,  1,  0, 
	}, 

	{
	-1,  3, -3,  1, 
	 3, -6,  3,  0, 
	-3,  0,  3,  0, 
	 1,  4,  1,  0, 
	}, 

	 {
	-1,  3, -2,  0, 
	 3, -6,  3,  0, 
	-3,  0,  3,  0, 
	 1,  4,  1,  0, 
	 }
};



inline void multiply(float a0, float a1, float a2, float a3, const int *mat, float *vec) {
	vec[0] = a0 * mat[ 0] + a1 * mat[ 1] + a2 * mat[ 2] + a3 * mat[ 3];
	vec[1] = a0 * mat[ 4] + a1 * mat[ 5] + a2 * mat[ 6] + a3 * mat[ 7];
	vec[2] = a0 * mat[ 8] + a1 * mat[ 9] + a2 * mat[10] + a3 * mat[11];
	vec[3] = a0 * mat[12] + a1 * mat[13] + a2 * mat[14] + a3 * mat[15];
}


inline float interpolate(float t, float scale_factor, float *vec) {
	float t2 = t * t;
	float t3 = t2 * t;

	return scale_factor * (t3 * vec[0] + t2 * vec[1] + t * vec[2] + vec[3]);
}


const float SplineStroke::init_threshold = 1.e-10f;


void SplineStroke::addBack(float x, float y) {
	float dx, dy, seg_length;

	//dx = x - xc;
	//dy = y - yc;
	dx = x - b_xc;
	dy = y - b_yc;

	if((seg_length = dx * dx + dy * dy) < threshold) return;

	ctrl_x.push_back(x);
	ctrl_y.push_back(y);
	num_ctrl_points++;

	squared_length += seg_length;
	length = sqrt(squared_length);

	//xc = x;
	//yc = y;
	b_xc = x;
	b_yc = y;
}


void SplineStroke::addFront(float x, float y) {
	float dx, dy, seg_length;

	//dx = x - xc;
	//dy = y - yc;
	dx = x - f_xc;
	dy = y - f_yc;

	if((seg_length = dx * dx + dy * dy) < threshold) return;

	ctrl_x.push_front(x);
	ctrl_y.push_front(y);
	num_ctrl_points++;

	squared_length += seg_length;
	length = sqrt(squared_length);

	//xc = x;
	//yc = y;
	f_xc = x;
	f_yc = y;
}


bool SplineStroke::tesselate(bool add_ends) {
	int idx = 0;
	int dp1 = density + 1;
	float hw = stroke_size * .5f;
	float tmp_squared_length = squared_length;
	float vec_x[4], vec_y[4];
	list<float>::iterator itr_x[4], itr_y[4];
	float tf_xc, tf_yc, tb_xc, tb_yc;

	tesselated = false;

	if(num_ctrl_points < 1) {
		return false;
	}


	// temporarily add two end points (we'll remove later)
	if(add_ends == true && hw * hw >= threshold) {
		float dx, dy;
		float t0, t1, tx, ty;
		list<float>::iterator itr0, itr1;

		tf_xc = f_xc;
		tf_yc = f_yc;
		tb_xc = b_xc;
		tb_yc = b_yc;

		switch(num_ctrl_points) {
			case 1:
				t0 = hw * cos(first_angle);
				t1 = hw * sin(first_angle);
				tx = ctrl_x.front();
				ty = ctrl_y.front();

				addFront(tx - t0, ty - t1);
				addBack(tx + t0, ty + t1);
				break;

			default:
				itr0 = ctrl_x.begin();
				itr1 = itr0;
				itr1++;
				dx = *itr1 - *itr0;
				itr0 = ctrl_y.begin();
				itr1 = itr0;
				itr1++;
				dy = *itr1 - *itr0;
				t0 = atan2(dy, dx);


				itr1 = ctrl_x.end();
				itr1--;
				itr0 = itr1;
				itr0--;
				dx = *itr1 - *itr0;
				itr1 = ctrl_y.end();
				itr1--;
				itr0 = itr1;
				itr0--;
				dy = *itr1 - *itr0;
				t1 = atan2(dy, dx);

				addFront(ctrl_x.front() - hw * cos(t0), ctrl_y.front() - hw * sin(t0));
				addBack(ctrl_x.back() + hw * cos(t1), ctrl_y.back() + hw * sin(t1));

				break;
		}
	}
	else {
		add_ends = false;
	}


	num_spline_points = num_ctrl_points + (num_ctrl_points - 1) * density;


	if(spline_points == NULL) {
		spline_points_capacity = num_spline_points;
		spline_points = new float[2 * spline_points_capacity];
	}
	else if(spline_points_capacity < num_spline_points) {
		delete[] spline_points;

		spline_points_capacity = num_spline_points;
		spline_points = new float[2 * spline_points_capacity];
	}


	switch(num_ctrl_points) {
		case 1:
			// just a single point
			spline_points[idx++] = *ctrl_x.begin();
			spline_points[idx++] = *ctrl_y.begin();
			break;

		case 2:
			// linear interpolation
			vec_x[0] = vec_x[1] = 0;
			vec_x[2] = *ctrl_x.rbegin() - *ctrl_x.begin();
			vec_x[3] = *ctrl_x.begin();
			
			vec_y[0] = vec_y[1] = 0;
			vec_y[2] = *ctrl_y.rbegin() - *ctrl_y.begin();
			vec_y[3] = *ctrl_y.begin();

			for(int j = 0; j <= dp1; j++) {
				spline_points[idx++] = interpolate(param_list[j], 1.f, vec_x);
				spline_points[idx++] = interpolate(param_list[j], 1.f, vec_y);
			}

			break;

		default:
			itr_x[3] = ctrl_x.begin();
			itr_x[0] = itr_x[3]++;
			itr_x[1] = itr_x[3]++;
			itr_x[2] = itr_x[3]++;

			itr_y[3] = ctrl_y.begin();
			itr_y[0] = itr_y[3]++;
			itr_y[1] = itr_y[3]++;
			itr_y[2] = itr_y[3]++;


			// first segment
			multiply(0, *itr_x[0], *itr_x[1], *itr_x[2], sc->m0, vec_x);
			multiply(0, *itr_y[0], *itr_y[1], *itr_y[2], sc->m0, vec_y);

			for(int j = 0; j < dp1; j++) {
				spline_points[idx++] = interpolate(param_list[j], sc->scale_factor, vec_x);
				spline_points[idx++] = interpolate(param_list[j], sc->scale_factor, vec_y);
			}


			// intermediate
			for(int i = 0; i < num_ctrl_points - 3; i++) {
				multiply(*itr_x[0], *itr_x[1], *itr_x[2], *itr_x[3], sc->mi, vec_x);
				multiply(*itr_y[0], *itr_y[1], *itr_y[2], *itr_y[3], sc->mi, vec_y);

				for(int j = 0; j < dp1; j++) {
					spline_points[idx++] = interpolate(param_list[j], sc->scale_factor, vec_x);
					spline_points[idx++] = interpolate(param_list[j], sc->scale_factor, vec_y);
				}

				itr_x[0]++; itr_x[1]++; itr_x[2]++; itr_x[3]++;
				itr_y[0]++; itr_y[1]++; itr_y[2]++; itr_y[3]++;
			}


			// last segment
			multiply(*itr_x[0], *itr_x[1], *itr_x[2], 0, sc->mn, vec_x);
			multiply(*itr_y[0], *itr_y[1], *itr_y[2], 0, sc->mn, vec_y);

			for(int j = 0; j <= dp1; j++) {
				spline_points[idx++] = interpolate(param_list[j], sc->scale_factor, vec_x);
				spline_points[idx++] = interpolate(param_list[j], sc->scale_factor, vec_y);
			}

			break;
	}


	// restore
	if(add_ends == true) {
		ctrl_x.pop_front();
		ctrl_x.pop_back();
		ctrl_y.pop_front();
		ctrl_y.pop_back();

		f_xc = tf_xc;
		f_yc = tf_yc;
		b_xc = tb_xc;
		b_yc = tb_yc;

		num_ctrl_points -= 2;
		squared_length = tmp_squared_length;
		length = sqrt(squared_length);
	}

	tesselated = true;

	return true;
}



void SplineStroke::setSplineType(SplineStroke::SPLINE_TYPE type) {
	tesselated = false;

	this->type = type;

	switch(type) {
		case CATMULL_ROM:
			sc = &SC_CATMULL_ROM;
			break;

		case UNIFORM_B_SPLINE:
			sc = &SC_UNIFORM_B_SPLINE;
			break;
	}
}


void SplineStroke::reset(float x0, float y0, float stroke_size, float stroke_color[4], float init_angle, int density) {
		this->stroke_size = stroke_size;
		this->density = density;
		length = 0;
		num_spline_points = 0;
		tesselated = false;
		num_ctrl_points = 1;
		threshold = init_threshold;
		first_angle = init_angle;
		id = -1;

		//xc = x0;
		//yc = y0;
		init_x = x0;
		init_y = y0;
		f_xc = x0;
		f_yc = y0;
		b_xc = x0;
		b_yc = y0;

		ctrl_x.clear();
		ctrl_y.clear();
		ctrl_x.push_back(x0);
		ctrl_y.push_back(y0);
		
		color[0] = stroke_color[0];
		color[1] = stroke_color[1];
		color[2] = stroke_color[2];
		color[3] = stroke_color[3];

		createParamList();
}
