#pragma once
#include "StdAfx.h"
#include "assert.h"
#include "TransitionalPaintingPreset.h"



TransitionalPaintingPreset::TransitionalPaintingPreset(PaintingPreset key) {
	keyframes[0] = key;
}


TransitionalPaintingPreset::~TransitionalPaintingPreset() {
}


PaintingPreset TransitionalPaintingPreset::get(int t) {
	float shift;
	int interpolation_func;
	int pt, nt;
	PaintingPreset *p, *n, r;
	map<int, PaintingPreset>::iterator itr;


	if(getSize() == 0) {
		return PRESET_NONE;
	}

	if((itr = keyframes.find(t)) != keyframes.end()) {
		return itr->second;
	}


	pt = getPrevKey(t);
	nt = getNextKey(t);

	assert(pt != nt);

	if(pt >= 0) {
		if(nt >= 0) {
			p = &(keyframes.find(pt)->second);
			n = &(keyframes.find(nt)->second);
			shift = ((float)t - pt) / (nt - pt);

			switch(key_interpolations.find(pt)->second) {
				case LINEAR:
				default:
					interpolation_func = 0;
					break;

				case SQUARE_ROOT1:
					interpolation_func = 1;
					break;

				case CUBE_ROOT1:
					interpolation_func = 2;
					break;

				case SQUARE_ROOT2:
					interpolation_func = 3;
					break;

				case CUBE_ROOT2:
					interpolation_func = 4;
					break;

				case EASE:
					interpolation_func = 5;
					break;
			}

			interpolatePaintingPreset(*p, *n, shift, r, interpolation_func);

			return r;
		}
		else {
			return keyframes.find(pt)->second;
		}
	}
	else {
		if(nt >= 0) {
			return keyframes.find(nt)->second;
		}
		else {
			assert("TransitionalPaintingPreset: issue on keyframe searching...");
			return PRESET_NONE;
		}
	}
}


size_t TransitionalPaintingPreset::getSize() {
	return keyframes.size();
}


int TransitionalPaintingPreset::getNextKey(int t) {
	map<int, PaintingPreset>::iterator it;

	if(keyframes.size() == 0) {
		return -1;
	}
	else {
		it = keyframes.upper_bound(t);

		if(it == keyframes.end()) {
			return -1;
		}
		else {
			return (*it).first;
		}
	}
}


int TransitionalPaintingPreset::getPrevKey(int t) {
	map<int, PaintingPreset>::iterator it;

	if(keyframes.size() == 0) {
		return -1;
	}
	else {
		it = keyframes.lower_bound(t);

		if(it == keyframes.begin()) {
			return -1;
		}
		else {
			it--;
			return (*it).first;
		}
	}
}


bool TransitionalPaintingPreset::set(PaintingPreset key, int t) {
	int p_key, n_key;
	map<int, PaintingPreset>::iterator itr;


	// return if invalid parameter is passed.
	if(t < 0) return false;

	// check whether any keyframe is set.
	itr = keyframes.find(t);

	// if no keyframe is found, copy interpolation function.
	if(itr == keyframes.end()) {
		p_key = getPrevKey(t);
		n_key = getNextKey(t);

		/**
		CString msg;
		msg.Format("Prev: %d, Next: %d\r\n", p_key, n_key);
		MessageBox(NULL, msg, _T(""), MB_OK);
		/**/

		// copy from the prev. keyframe if in-between two keyframes or set as LINEAR otherwise.
		if(p_key != -1) {
			if(n_key != -1) {
				key_interpolations[t] = key_interpolations[p_key];
			}
			else {
				key_interpolations[p_key] = LINEAR;
				key_interpolations[t] = UNDEFINED;
			}
		}
		else {
			if(n_key != -1) {
				key_interpolations[t] = LINEAR;
			}
			else {
				key_interpolations[t] = UNDEFINED;
			}
		}
	}

	// set (or replace) a keyframe
	keyframes[t] = key;

	return true;
}


bool TransitionalPaintingPreset::remove(int t) {
	map<int, PaintingPreset>::iterator it;
	map<int, INTERPOLATION_TYPE>::reverse_iterator rit_inter_func;

	if(t < 0 || keyframes.size() <= 0) return false;

	if((it = keyframes.find(t)) == keyframes.end()) {
		return false;
	}
	else {
		keyframes.erase(it);
		
		key_interpolations.erase(key_interpolations.find(t));

		// make sure the interpolation func of the last element is always undefined.
		if((rit_inter_func = key_interpolations.rbegin()) != key_interpolations.rend()) {
			rit_inter_func->second = UNDEFINED;
		}

		return true;
	}
}


void TransitionalPaintingPreset::copy(const TransitionalPaintingPreset &src) {
	keyframes = src.keyframes;
	key_interpolations = src.key_interpolations;
}


void TransitionalPaintingPreset::getAllKeys(vector< pair<int, PaintingPreset> > &keys) {
	map<int, PaintingPreset>::iterator key_itr, key_itr_end;

	key_itr = keyframes.begin();
	key_itr_end = keyframes.end();

	keys.resize(0);

	for(; key_itr != key_itr_end; key_itr++) {
		keys.push_back(*key_itr);		
	}
}


void TransitionalPaintingPreset::setAllKeys(const vector< pair<int, PaintingPreset> > &keys) {
	size_t key_size;


	keyframes.clear();
	key_interpolations.clear();


	key_size = keys.size();

	for(size_t i = 0; i < key_size; i++) {
		keyframes[keys[i].first] = keys[i].second;

		if(i != key_size - 1) {
			key_interpolations[keys[i].first] = LINEAR;
		}
		else {
			key_interpolations[keys[i].first] = UNDEFINED;
		}
	}
}



void TransitionalPaintingPreset::setInterpolationFunc(TransitionalPaintingPreset::INTERPOLATION_TYPE func_type, int t) {
	int p_key;

	if(key_interpolations.find(t) != key_interpolations.end()) {
		if(key_interpolations[t] != UNDEFINED) {
			key_interpolations[t] = func_type;
		}
	}
	else if((p_key = getPrevKey(t)) != -1) {
		if(key_interpolations[p_key] != UNDEFINED) {
			key_interpolations[p_key] = func_type;
		}
	}
	else {
		return;
	}
}


TransitionalPaintingPreset::INTERPOLATION_TYPE TransitionalPaintingPreset::getInterpolationFunc(int t) {
	int p_key;

	if(key_interpolations.find(t) != key_interpolations.end()) {
		return key_interpolations[t];
	}
	else if((p_key = getPrevKey(t)) != -1) {
		return key_interpolations[p_key];
	}
	else {
		return UNDEFINED;
	}
}


void TransitionalPaintingPreset::getAllInterpolationFuncs(vector< pair<int, INTERPOLATION_TYPE> > &functions) {
	map<int, INTERPOLATION_TYPE>::iterator func_itr, func_itr_end;

	func_itr = key_interpolations.begin();
	func_itr_end = key_interpolations.end();

	functions.resize(0);

	for(; func_itr != func_itr_end; func_itr++) {
		functions.push_back(*func_itr);		
	}
}


void TransitionalPaintingPreset::setAllInterpolationFuncs(const vector< pair<int, INTERPOLATION_TYPE> > &functions) {
	size_t func_size;


	func_size = functions.size();

	// check non-empty
	if(func_size == 0) {
		return;
	}

	// check element size
	if(func_size != keyframes.size()) {
		return;
	}

	// should have the same size...
	if(func_size != key_interpolations.size()) {
		return;
	}


	for(size_t i = 0; i < func_size; i++) {
		if(key_interpolations.find(functions[i].first) != key_interpolations.end()) {
			key_interpolations[functions[i].first] = (functions[i].second != UNDEFINED) ? functions[i].second : LINEAR;
		}
	}

	// ensure the last element is always undefined...
	key_interpolations.rbegin()->second = UNDEFINED;
}

